# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Code style validation wrapper for screwdrivercd

This wrapper runs the pycodestyle validation tool.
"""
# The logging_basicConfig has to be run before other imports because some modules we use log output on import
# pylint: disable=wrong-import-order, wrong-import-position
from ..screwdriver.environment import logging_basicConfig
logging_basicConfig(check_prefix='PACKAGE_QUALITY_CHECK')

import logging
import os
import subprocess  # nosec
import sys
from termcolor import colored

from typing import List

from ..utility.environment import env_int, env_bool, interpreter_bin_command, standard_directories
from ..utility.run import run_and_log_output

logger_name = 'validate_package_quality' if __name__ == '__main__' else __name__
logger = logging.getLogger(logger_name)


def validate_package_quality(package_dir: str='') -> int:
    """
    Run the pyroma tool to validate the quality of packages in the package directory

    Parameters
    ----------
    package_dir: str
        The package directory containing the packages to check

    Returns
    -------
    int:
        Return code from the validation command
    """
    directories = standard_directories('package_quality_validation')
    if not package_dir:  # pragma: no cover
        package_dir = directories['packages']

    pyroma_min_score = env_int('PYROMA_MIN_SCORE', 8)
    fail_if_no_packages = env_bool('VALIDATE_PACKAGE_QUALITY_FAIL_MISSING', False)

    if not os.path.exists(package_dir):  # pragma: no cover
        print(f'Package directory {package_dir!r} is not present, no packages to validate')
        if fail_if_no_packages:
            return 1
        else:
            return 0
    return_codes: List[int] = []
    checked_package = False
    for package in os.listdir(package_dir):
        if package.endswith('.whl'):
            continue
        print(f'Checking package: {package}', flush=True, end='')
        package_filename = os.path.join(package_dir, package)
        report_filename = f'{directories["reports"]}/{package}.log'
        command = [interpreter_bin_command('pyroma'), f'--min={pyroma_min_score}', package_filename]
        checked_package = True
        try:
            run_and_log_output(command, logfile=report_filename)
            return_codes.append(0)
        except subprocess.CalledProcessError as error:
            print(f' ... {colored("Failed", "red")}', flush=True)
            return_codes.append(error.returncode)
    if fail_if_no_packages and not checked_package:
        print(f' ... {colored("No source packages to validate", "red")}')
        return 1
    return sum(return_codes)


def main() -> int:
    """
    Pyroma utility running command line utility entry point

    Returns
    -------
    int:
        Returncode from running the check
    """
    return validate_package_quality()  # pragma: no cover


if __name__ == '__main__':
    sys.exit(main())



"""
Sphinx documentation generation plugin
"""
import os
from ..plugin import DocumentationPlugin
from ...utility.environment import interpreter_bin_command


class SphinxDocumentationPlugin(DocumentationPlugin):
    """
    screwdrivercd.documentation plugin for sphinx documentation
    """
    name = 'sphinx'
    build_dest = 'build/sphinx/html'
    builder = 'html'

    def __init__(self, *args, **kwargs):  # pragma: no cover
        super().__init__(*args, **kwargs)
        self.source_dir = os.path.join(self.source_dir, 'doc/source')
        self.build_output_dir = 'build/sphinx/html'
        self.builder = os.environ.get('DOCUMENTATION_SPHINX_BUILDER', self.builder)
        self.build_command = [interpreter_bin_command('sphinx-build'), '-b', self.builder, self.source_dir, self.build_dest]

    @property
    def documentation_is_present(self) -> bool:
        if os.path.exists('doc/source/conf.py'):
            return True
        return False



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Screwdriver Utility ContextManagers
"""
import logging
import os
import signal
from contextlib import contextmanager, ContextDecorator
from datetime import timedelta
from tempfile import TemporaryDirectory
from typing import Optional, Union

from .exceptions import TimeoutError


logger = logging.getLogger(__name__)


@contextmanager
def working_dir(new_path):
    """
    A context manager that changes to the new_path directory and
    returns to the current working directory when it completes.
    """
    old_dir = os.getcwd()
    os.chdir(new_path)
    try:
        yield new_path
    finally:
        os.chdir(old_dir)


@contextmanager
def revert_file(filename):  # pragma: no cover
    """
    A context manager that reverts a file's contents.
    """
    original_data = None
    exists = os.path.exists(filename)
    if exists:
        with open(filename, 'rb') as file_handle:
            original_data = file_handle.read()

    try:
        yield filename
    finally:
        if exists:
            if original_data:
                with open(filename, 'wb') as file_handle:
                    file_handle.write(original_data)
        else:
            if os.path.exists(filename):
                os.remove(filename)


@contextmanager
def InTemporaryDirectory(suffix=None, prefix=None, dir=None):
    """
    A context manager that creates a temporary directory and switches into it
    """
    with TemporaryDirectory(suffix=suffix, prefix=prefix, dir=dir) as tempdir:
        with working_dir(tempdir):
            yield tempdir


class Timeout(ContextDecorator):
    """
    A contextmanager that will timeout after a specific datetime.timedelta
    """
    use_alarm: Union[bool, None] = False
    _old_handler = None

    def __init__(self, timeout: Optional[timedelta]=None, use_alarm: Union[bool, None]=None):
        self.timeout = timeout
        self.use_alarm = use_alarm
        if use_alarm is None:
            # use_alarm wasn't specified, set the value based on support for signal.setitimer()
            if hasattr(signal, 'setitimer'):
                self.use_alarm = False
            else:  # pragma: no cover
                self.use_alarm = True

    def _timeout_handler(self, signum, frame):
        raise TimeoutError(f'Timeout after {self.timeout}')

    def __enter__(self):
        if self.timeout:
            if self.use_alarm:
                self._old_handler = signal.signal(signal.SIGALRM, self._timeout_handler)
                signal.alarm(self.timeout.seconds)
            else:
                self._old_handler = signal.signal(signal.SIGALRM, self._timeout_handler)
                signal.setitimer(signal.ITIMER_REAL, float(self.timeout.microseconds) / 1000000)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.timeout:
            if self.use_alarm:
                signal.alarm(0)
            else:
                signal.setitimer(signal.ITIMER_REAL, 0)
        if self._old_handler:
            signal.signal(signal.SIGALRM, self._old_handler)



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
import os
import unittest
import sys
import tempfile
from screwdrivercd.documentation import cli


class CliTestCase(unittest.TestCase):
    used_env_vars = ['DOCUMENTATION_DEBUG', 'DOCUMENTATION_FORMATS', 'DOCUMENTATION_PUBLISH']
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._argv = sys.argv
        self._cwd = os.getcwd()

    def setUp(self):
        self.clear_used_env_keys()

    def tearDown(self):
        sys.argv = self._argv
        os.chdir(self._cwd)
        self.clear_used_env_keys()

    def clear_used_env_keys(self):
        for varname in self.used_env_vars:
            if varname in os.environ.keys():
                del os.environ[varname]

    def test__default__base(self):
        os.environ['DOCUMENTATION_FORMATS'] = 'base'
        sys.argv = ['sdv4_documentation']
        cli.main()

    def test__build__mkdocs__none(self):
        os.environ['DOCUMENTATION_FORMATS'] = 'mkdocs'
        os.environ['DOCUMENTATION_PUBLISH'] = 'False'
        sys.argv = ['sdv4_documentation']
        with tempfile.TemporaryDirectory() as tempdir:
            os.chdir(tempdir)
            cli.main()

    def test__build__mkdocs(self):
        os.environ['DOCUMENTATION_FORMATS'] = 'mkdocs'
        os.environ['DOCUMENTATION_PUBLISH'] = 'False'
        sys.argv = ['sdv4_documentation']
        with tempfile.TemporaryDirectory() as tempdir:
            os.chdir(tempdir)
            os.makedirs('docs')
            with open('mkdocs.yml', 'w') as fh:
                fh.write('site_name: test\n')

            with open('docs/index.md', 'w') as fh:
                fh.write('# Title\n')
            cli.main()

    def test__build__mkdocs_venv(self):
        os.environ['DOCUMENTATION_FORMATS'] = 'mkdocs_venv'
        os.environ['DOCUMENTATION_PUBLISH'] = 'False'
        sys.argv = ['sdv4_documentation']
        with tempfile.TemporaryDirectory() as tempdir:
            os.chdir(tempdir)
            os.makedirs('docs')
            with open('mkdocs.yml', 'w') as fh:
                fh.write('site_name: test\n')

            with open('docs/index.md', 'w') as fh:
                fh.write('# Title\n')
            result = cli.main()
        self.assertEqual(result, 0)

    def test__build__mkdocs_venv__requirements_missing__fail(self):
        os.environ['DOCUMENTATION_FORMATS'] = 'mkdocs_venv'
        os.environ['DOCUMENTATION_PUBLISH'] = 'False'
        sys.argv = ['sdv4_documentation']
        with tempfile.TemporaryDirectory() as tempdir:
            os.chdir(tempdir)
            os.makedirs('docs')
            with open('mkdocs.yml', 'w') as fh:
                fh.write('site_name: test\nmarkdown_extensions:\n    - pymdownx.arithmatex\n')

            with open('docs/index.md', 'w') as fh:
                fh.write('# Title\n')

            with open('documentation_mkdocs_requirements.txt', 'w') as fh:
                fh.write('mkdocs\n')

            result = cli.main()
        self.assertNotEqual(result, 0)

    def test__build__mkdocs_venv__requirements_present(self):
        os.environ['DOCUMENTATION_FORMATS'] = 'mkdocs_venv'
        os.environ['DOCUMENTATION_PUBLISH'] = 'False'
        sys.argv = ['sdv4_documentation']
        with tempfile.TemporaryDirectory() as tempdir:
            os.chdir(tempdir)
            os.makedirs('docs')
            with open('mkdocs.yml', 'w') as fh:
                fh.write('site_name: test\nmarkdown_extensions:\n    - pymdownx.arithmatex\n')

            with open('docs/index.md', 'w') as fh:
                fh.write('# Title\n')

            with open('documentation_mkdocs_requirements.txt', 'w') as fh:
                fh.write('mkdocs\npymdown-extensions\n')

            result = cli.main()
        self.assertEqual(result, 0)


    def test__build__mkdocs__docsdir(self):
        os.environ['DOCUMENTATION_FORMATS'] = 'mkdocs'
        os.environ['DOCUMENTATION_PUBLISH'] = 'False'
        sys.argv = ['sdv4_documentation']
        with tempfile.TemporaryDirectory() as tempdir:
            os.chdir(tempdir)
            os.makedirs('docs')
            with open('docs/mkdocs.yml', 'w') as fh:
                fh.write('site_name: test\n')

            with open('docs/index.md', 'w') as fh:
                fh.write('# Title\n')
            cli.main()

    def test__build__mkdocs__fail(self):
        os.environ['DOCUMENTATION_FORMATS'] = 'mkdocs'
        os.environ['DOCUMENTATION_PUBLISH'] = 'False'
        sys.argv = ['sdv4_documentation']
        with tempfile.TemporaryDirectory() as tempdir:
            os.chdir(tempdir)
            os.makedirs('docs')
            with open('mkdocs.yml', 'w') as fh:
                fh.write('site_name: test\nstrict: true\nnav:\n    - foo: bar.md\n')

            with open('docs/index.md', 'w') as fh:
                fh.write('# Title\n[bad_link](bar.md)\n')
            result = cli.main()
            self.assertNotEqual(result, 0)

    def test__build__sphinx__none(self):
        os.environ['DOCUMENTATION_FORMATS'] = 'sphinx'
        os.environ['DOCUMENTATION_PUBLISH'] = 'False'
        sys.argv = ['sdv4_documentation']
        with tempfile.TemporaryDirectory() as tempdir:
            os.chdir(tempdir)
            cli.main()

    def test__build__sphinx(self):
        os.environ['DOCUMENTATION_DEBUG'] = 'True'
        os.environ['DOCUMENTATION_FORMATS'] = 'sphinx'
        os.environ['DOCUMENTATION_PUBLISH'] = 'False'
        sys.argv = ['sdv4_documentation']
        with tempfile.TemporaryDirectory() as tempdir:
            os.chdir(tempdir)
            os.makedirs('doc/source')
            with open('doc/source/conf.py', 'w') as fh:
                fh.write("""# -*- coding: utf-8 -*-
import os
extensions = []
templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'
project = 'foo'
copyright = ''
version = '0.0.0'
release = version
""")
            with open('doc/source/index.rst', 'w') as fh:
                fh.write('''Title\n-----\n\n''')

            cli.main()

    def test__publish__mkdocs__fail(self):
        os.environ['DOCUMENTATION_FORMATS'] = 'mkdocs'
        os.environ['DOCUMENTATION_PUBLISH'] = 'True'
        sys.argv = ['sdv4_documentation']
        with tempfile.TemporaryDirectory() as tempdir:
            os.chdir(tempdir)
            os.makedirs('docs')
            with open('mkdocs.yml', 'w') as fh:
                fh.write('site_name: test\nstrict: true\nnav:\n    - foo: bar.md\n')

            with open('docs/index.md', 'w') as fh:
                fh.write('# Title\n[bad_link](bar.md)\n')
            result = cli.main()
            self.assertNotEqual(result, 0)


# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""Exception classes"""


class VersionError(Exception):
    """An error has occurred generating a version"""



#!/usr/bin/env python
# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
import os
import unittest
import unittest.mock

import distro
from screwdrivercd.installdeps.installer import Installer
from screwdrivercd.utility.contextmanagers import InTemporaryDirectory

from . import ScrewdriverTestCase


CONFIG_FILE = 'pyproject.toml'
TEST_CONFIG = f'''[build-system]
# Minimum requirements for the build system to execute.
requires = ["setuptools", "wheel"]  # PEP 508 specifications.

[tool.sdv4_installdeps]
install = ['echo']

[tool.sdv4_installdeps.echo]
deps = ['python3', 'foo;distro_version=="{distro.version()}"', 'bar;distro_version!="{distro.version()}"']
'''

TEST_CONFIG_NODEPS = f'''[build-system]
# Minimum requirements for the build system to execute.
requires = ["setuptools", "wheel"]  # PEP 508 specifications.

[tool.sdv4_installdeps]
install = ['echo']

[tool.sdv4_installdeps.echo]
deps = []
'''

class TestEcho(ScrewdriverTestCase):

    installer_class = Installer

    def setUp(self):
        super().setUp()
        with open(CONFIG_FILE, 'w') as config_handle:
            config_handle.write(TEST_CONFIG)
        self.installer = Installer(bin_dir='/bin')

    def test__echo__has_dependencies(self):
        result = self.installer.has_dependencies
        self.assertEqual(result, True)

    def test__echo__is_supported(self):
        result = self.installer.is_supported
        self.assertEqual(result, True)

    def test__echo__install__nodeps(self):
        with open(CONFIG_FILE, 'w') as config_handle:
            config_handle.write(TEST_CONFIG_NODEPS)
        self.installer = Installer(bin_dir='/bin')
        result = self.installer.install_dependencies()
        self.assertListEqual(result, [])

    def test__echo__install__default_python(self):
        result = self.installer.install_dependencies()
        self.assertListEqual(result, ['python3', 'foo'])

    def test__echo__install__default_python__dry_run(self):
        installer = self.installer_class(dry_run=True)
        result = installer.install_dependencies()
        self.assertListEqual(result, ['python3', 'foo'])

    def test__echo__invalid_dependencies(self):
        with unittest.mock.patch.object(Installer, 'validate_dependency', return_value=False) as mock_method:
            result = self.installer.install_dependencies()
            self.assertListEqual(result, [])

    def test__echo__install__invalid_deps(self):
        with unittest.mock.patch.object(Installer, 'validate_dependency', return_value=False) as mock_method:
            result = self.installer.install_dependencies()
            self.assertListEqual(result, [])

    def test__echo__install__invalid_deps__log_output(self):
        with unittest.mock.patch.object(Installer, 'validate_dependency', return_value=False) as mock_method:
            self.installer.print_output = False
            result = self.installer.install_dependencies()
            self.assertListEqual(result, [])

    def test__echo__install__invalid_deps__exit_on_missing(self):
        with unittest.mock.patch.object(Installer, 'validate_dependency', return_value=False) as mock_method:
            self.installer.exit_on_missing = True
            result = self.installer.install_dependencies()
            self.assertListEqual(result, [])

    def test__determine_bin_directory__install_command_path(self):
        self.installer.install_command[0] = 'echo'
        self.installer.bin_dir = None
        self.installer.install_command_path = ['/bin']
        self.installer.determine_bin_directory()

    def test__determine_bin_directory__system_path(self):
        self.installer.install_command[0] = 'echo'
        self.installer.bin_dir = None
        self.installer.install_command_path = []
        self.installer.use_system_path = True
        self.installer.determine_bin_directory()

    def test_find_install_command__bin_dir(self):
        self.installer.install_command[0] = 'echo'
        self.installer.bin_dir = '/bin'
        self.installer.install_command_path = []
        self.installer.find_install_command()
        self.assertEqual(self.installer.install_command[0], '/bin/echo')

    def test_find_install_command__system_dir(self):
        self.installer.install_command[0] = 'echo'
        self.installer.bin_dir = None
        self.installer.install_command_path = []
        self.installer.use_system_path = True
        self.installer.find_install_command()



# This is based on the pep508-parser grammer at https://github.com/NFJones/pep508-parser
# MIT Licensed
"""
Requirements parsing functions
"""
import logging
import os
import platform
import sys
from typing import Dict

import distro
import parsley

from packaging.version import Version, parse

logger = logging.getLogger(__name__)


GRAMMAR: str = """
    wsp            = ' ' | '\t'
    version_cmp    = wsp* <'<=' | '<' | '!=' | '==' | '>=' | '>' | '~=' | '==='>
    version        = wsp* <( letterOrDigit | '-' | '_' | '.' | '*' | '+' | '!' )+>
    version_one    = version_cmp:op version:v wsp* -> (op, v)
    version_many   = version_one:v1 (wsp* ',' version_one)*:v2 -> [v1] + v2
    versionspec    = ('(' version_many:v ')' ->v) | version_many
    urlspec        = '@' wsp* <URI_reference>
    marker_op      = version_cmp | (wsp* 'in') | (wsp* 'not' wsp+ 'in')
    python_str_c   = (wsp | letter | digit | '(' | ')' | '.' | '{' | '}' |
                      '-' | '_' | '*' | '#' | ':' | ';' | ',' | '/' | '?' |
                      '[' | ']' | '!' | '~' | '`' | '@' | '$' | '%' | '^' |
                      '&' | '=' | '+' | '|' | '<' | '>' )
    dquote         = '"'
    squote         = '\\''
    python_str     = (squote <(python_str_c | dquote)*>:s squote |
                      dquote <(python_str_c | squote)*>:s dquote) -> s
    env_var        = ('python_version' | 'python_full_version' |
                      'os_name' | 'sys_platform' | 'platform_release' |
                      'platform_system' | 'platform_version' |
                      'platform_machine' | 'platform_python_implementation' |
                      'implementation_name' | 'implementation_version' |
                      'distro_codename' | 'distro_name' | 'distro_id' |
                      'distro_like' | 'distro_version' |
                      'extra' # ONLY when defined by a containing layer
                      ):varname -> lookup(varname)
    marker_var     = wsp* (env_var | python_str)
    marker_expr    = marker_var:l marker_op:o marker_var:r -> (o, l, r)
                   | wsp* '(' marker:m wsp* ')' -> m
    marker_and     = marker_expr:l wsp* 'and' marker_expr:r -> ('and', l, r)
                   | marker_expr:m -> m
    marker_or      = marker_and:l wsp* 'or' marker_and:r -> ('or', l, r)
                       | marker_and:m -> m
    marker         = marker_or
    quoted_marker  = ';' wsp* marker
    identifier_end = letterOrDigit | (('-' | '_' | '.' )* letterOrDigit)
    identifier     = < letterOrDigit identifier_end* >
    name           = identifier
    extras_list    = identifier:i (wsp* ',' wsp* identifier)*:ids -> [i] + ids
    extras         = '[' wsp* extras_list?:e wsp* ']' -> e
    name_req       = (name:n wsp* extras?:e wsp* versionspec?:v wsp* quoted_marker?:m
                      -> (n, e or [], v or [], m))
    url_req        = (name:n wsp* extras?:e wsp* urlspec:v (wsp+ | end) quoted_marker?:m
                      -> (n, e or [], v, m))
    specification  = wsp* ( url_req | name_req ):s wsp* -> s
    # The result is a tuple - name, list-of-extras,
    # list-of-version-constraints-or-a-url, marker-ast or None


    URI_reference = <URI | relative_ref>
    URI           = scheme ':' hier_part ('?' query )? ( '#' fragment)?
    hier_part     = ('//' authority path_abempty) | path_absolute | path_rootless | path_empty
    absolute_URI  = scheme ':' hier_part ( '?' query )?
    relative_ref  = relative_part ( '?' query )? ( '#' fragment )?
    relative_part = '//' authority path_abempty | path_absolute | path_noscheme | path_empty
    scheme        = letter ( letter | digit | '+' | '-' | '.')*
    authority     = ( userinfo '@' )? host ( ':' port )?
    userinfo      = ( unreserved | pct_encoded | sub_delims | ':')*
    host          = IP_literal | IPv4address | reg_name
    port          = digit*
    IP_literal    = '[' ( IPv6address | IPvFuture) ']'
    IPvFuture     = 'v' hexdig+ '.' ( unreserved | sub_delims | ':')+
    IPv6address   = (
                      ( h16 ':'){6} ls32
                      | '::' ( h16 ':'){5} ls32
                      | ( h16 )?  '::' ( h16 ':'){4} ls32
                      | ( ( h16 ':')? h16 )? '::' ( h16 ':'){3} ls32
                      | ( ( h16 ':'){0,2} h16 )? '::' ( h16 ':'){2} ls32
                      | ( ( h16 ':'){0,3} h16 )? '::' h16 ':' ls32
                      | ( ( h16 ':'){0,4} h16 )? '::' ls32
                      | ( ( h16 ':'){0,5} h16 )? '::' h16
                      | ( ( h16 ':'){0,6} h16 )? '::' )
    h16           = hexdig{1,4}
    ls32          = ( h16 ':' h16) | IPv4address
    IPv4address   = dec_octet '.' dec_octet '.' dec_octet '.' dec_octet
    nz            = ~'0' digit
    dec_octet     = (
                      digit # 0-9
                      | nz digit # 10-99
                      | '1' digit{2} # 100-199
                      | '2' ('0' | '1' | '2' | '3' | '4') digit # 200-249
                      | '25' ('0' | '1' | '2' | '3' | '4' | '5') )# %250-255
    reg_name = ( unreserved | pct_encoded | sub_delims)*
    path = (
            path_abempty # begins with '/' or is empty
            | path_absolute # begins with '/' but not '//'
            | path_noscheme # begins with a non-colon segment
            | path_rootless # begins with a segment
            | path_empty ) # zero characters
    path_abempty  = ( '/' segment)*
    path_absolute = '/' ( segment_nz ( '/' segment)* )?
    path_noscheme = segment_nz_nc ( '/' segment)*
    path_rootless = segment_nz ( '/' segment)*
    path_empty    = pchar{0}
    segment       = pchar*
    segment_nz    = pchar+
    segment_nz_nc = ( unreserved | pct_encoded | sub_delims | '@')+
                    # non-zero-length segment without any colon ':'
    pchar         = unreserved | pct_encoded | sub_delims | ':' | '@'
    query         = ( pchar | '/' | '?')*
    fragment      = ( pchar | '/' | '?')*
    pct_encoded   = '%' hexdig
    unreserved    = letter | digit | '-' | '.' | '_' | '~'
    reserved      = gen_delims | sub_delims
    gen_delims    = ':' | '/' | '?' | '#' | '(' | ')?' | '@'
    sub_delims    = '!' | '$' | '&' | '\\'' | '(' | ')' | '*' | '+' | ',' | ';' | '='
    hexdig        = digit | 'a' | 'A' | 'b' | 'B' | 'c' | 'C' | 'd' | 'D' | 'e' | 'E' | 'f' | 'F'
"""


class Requirement():
    """
    Requirement parser
    """
    _parser = None
    _lookup: Dict[str, str] = dict(
        distro_codename=distro.codename(),
        distro_id=distro.id(),
        distro_like=distro.like(),
        distro_name=distro.name(),
        distro_version=distro.version(),
        os_name=os.name,
        platform_machine=platform.machine(),
        platform_python_implementation=platform.python_implementation(),
        platform_release=platform.release(),
        platform_system=platform.system(),
        platform_version=platform.version(),
        platform_full_version=platform.python_version(),
        python_version=platform.python_version()[:3],
        sys_platform=sys.platform
    )

    def __init__(self, s):
        self._parser = parsley.makeGrammar(GRAMMAR, {"lookup": self._lookup.__getitem__})
        self._parsed_requirement = self._parser(s).specification()

    @property
    def name(self):
        """
        Get the name part of the requirement
        """
        return self._parsed_requirement[0]

    @property
    def extra(self):
        """
        Get the extras from the requirement
        """
        return self._parsed_requirement[1]

    @property
    def version_evals(self):
        """
        Return the version evaluations for the requirement if present
        """
        if not self._parsed_requirement[2]:
            return []
        return self._parsed_requirement[2]

    @property
    def env_evals(self):
        """
        Return any environment evaluations for the requirement
        """
        if not self._parsed_requirement[3]:
            return []
        return self._parsed_requirement[3]

    @property
    def env_matches(self):
        """
        Determine if the current env matches the requirement

        Returns
        -------
        True if it matches
        """
        env_evals = list(self.env_evals)
        env_evals = self.evaluate_matches(env_evals)
        if len(env_evals) == 3:
            env_evals = [self.evaluate(*env_evals)]
        return False not in env_evals

    def evaluate_matches(self, env_evals):
        """
        Evaluate all of the statements and return a list of the results
        Parameters
        ----------
        env_evals: evals to be evaluated

        Returns
        -------
        Evaluates all the statements and return the ones that match
        """
        for element_num in range(len(env_evals)):  # pylint: disable=consider-using-enumerate
            entry = env_evals[element_num]
            if isinstance(entry, str):
                continue
            operation, val1, val2 = entry
            env_evals[element_num] = self.evaluate(operation, val1, val2)
        return env_evals

    def evaluate(self, operation, val1, val2):
        """Evaluate"""
        if isinstance(val1, tuple):
            val1 = self.evaluate(*val1)
        if isinstance(val2, tuple):
            val2 = self.evaluate(*val2)

        if isinstance(val1, str):
            val1 = parse(val1)
        if isinstance(val2, str):
            val2 = parse(val2)

        result = None
        if operation == '>':
            result = val1 > val2
        elif operation == '<':
            result = val1 < val2
        elif operation == '>=':
            result = val1 >= val2
        elif operation == '<=':
            result = val1 <= val2
        elif operation == '==':
            result = val1 == val2
        elif operation in ['!', '!=']:
            result = val1 != val2
        elif operation == 'and':
            result = val1 and val2
        elif operation == 'or':
            result = val1 or val2
        else:
            logger.error(f'Invalid operation {operation}')
        return result



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Documentation generator plugin classes
"""
import importlib.metadata
import logging
import os
import subprocess  # nosec - All subprocess calls use full path
import sys
import tempfile

from typing import List

from termcolor import colored

from .exceptions import DocBuildError, DocPublishError
from .utility import clean_directory, copy_contents
from ..changelog.generate import write_changelog
from ..utility.environment import interpreter_bin_command, standard_directories


logger = logging.getLogger(__name__)


class DocumentationPlugin:
    """
    screwdrivercd.documentation generation/publication plugin
    """
    name: str = 'base'
    build_command: List[str] = [interpreter_bin_command(), '-c', 'print("building")']
    build_log_filename: str = ''
    build_output_dir = ''
    build_dest: str = ''
    publish_branch: str = 'gh-pages'
    publish_log_filename: str = ''
    git_command_timeout: int = 300
    _clone_dir = ''
    _clone_url = ''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        directories = standard_directories('documentation')
        self.interpreter_bin_dir = os.path.dirname(sys.executable)
        self.build_dir = directories['documentation']
        self.log_dir = directories['logs']
        self.build_log_filename = os.path.join(self.log_dir, f'{self.name}.build.log')
        self.publish_log_filename = os.path.join(self.log_dir, f'{self.name}.publish.log')
        self.build_dest = os.path.join(self.build_dir, self.build_output_dir)
        self.source_dir = os.getcwd()
        try:
            self.git_command_timeout = int(os.environ.get("DOCUMENTATION_GIT_TIMEOUT", str(self.git_command_timeout)))
        except ValueError:
            logger.warning('Value {os.environ.get("DOCUMENTATION_GIT_TIMEOUT", str(self.git_command_timeout))} for the DOCUMENTATION_GIT_TIMEOUT setting is invalid')
    @property
    def clone_dir(self) -> str:
        """
        Determine the git clone url for the current repo

        Returns
        -------
        str:
            Clone url, will be an empty string if it cannot be found
        """
        if not self._clone_dir:  # pragma: no cover
            self._clone_dir = self.get_clone_dir()
        return self._clone_dir

    @property
    def clone_url(self) -> str:
        """
        Determine the git clone url for the current repo

        Returns
        -------
        str:
            Clone url, will be an empty string if it cannot be found
        """
        if not self._clone_url:  # pragma: no cover
            self._clone_url = self.get_clone_url()
        return self._clone_url

    @property
    def documentation_is_present(self) -> bool:  # pragma: no cover
        """
        Returns
        =======
        bool:
            Check that the documentation source is present, Returns True if it is, False otherwise
        """
        return False

    def _run_command(self, command: List, log_filename: str, timeout: int = 0):
        """

        Parameters
        ----------
        command: List
            Parsed command to run

        log_filename: str
            Filename to write the log entry to
        """
        log_dir = os.path.dirname(log_filename)
        if log_dir:
            os.makedirs(log_dir, exist_ok=True)

        logger.debug(f'Running command {command!r}, logging to {log_filename!r}')
        error_message = None
        with open(log_filename, 'ab') as log_file:
            try:
                if timeout:  # pragma: no cover
                    subprocess.check_call(command, stdout=log_file, stderr=subprocess.STDOUT, timeout=timeout)  # nosec - All subprocess calls use full path
                else:
                    subprocess.check_call(command, stdout=log_file, stderr=subprocess.STDOUT)  # nosec - All subprocess calls use full path
            except (FileNotFoundError, subprocess.CalledProcessError,) as error:
                error_message = f'Command {" ".join(self.build_command)!r} failed {error}'
            except subprocess.TimeoutExpired as error:  # pragma: no cover
                error_message = f'Command {" ".join(self.build_command)!r} timeout after {timeout} seconds {error}'
        if error_message:
            raise DocBuildError(error_message)

    def _log_message(self, message: str, log_filename: str = '', end: str = '\n'):
        if not log_filename:
            log_filename = self.build_log_filename
        log_dir = os.path.dirname(log_filename)
        if log_dir:
            os.makedirs(log_dir, exist_ok=True)
        with open(log_filename, 'ab') as log_file:
            log_file.write(message.encode())
            if end:
                log_file.write(end.encode())

    def remove_build_log(self):
        """Remove the build log"""
        if os.path.exists(self.build_log_filename):
            os.remove(self.build_log_filename)

    def remove_publish_log(self):
        """Remove the publish log"""
        if os.path.exists(self.publish_log_filename):
            os.remove(self.publish_log_filename)

    def clean_directory(self, directory_name: str):
        """
        Clean out the specified directory

        Parameters
        ----------
        directory_name : str
            Directory to empty
        """
        self._log_message(f'\n- Cleaning out the old document contents', self.publish_log_filename)
        clean_directory(directory_name=directory_name)

    def copy_contents(self, src: str, dest: str):
        """
        Copy the contents of the source directory to the destination directory

        Parameters
        ----------
        src: str
            The source directory

        dest: str
            The destination directory
        """
        self._log_message(f'\n- Copying files from {src} -> {dest}', self.publish_log_filename)
        copy_contents(src, dest)

    def clone_documentation_branch(self):  # pragma: no cover
        """
        Clone and checkout the current documentation branch
        """
        self._log_message(f'\n- Cloning the {self.publish_branch!r} of the {self.clone_url!r} repo', self.publish_log_filename)
        if os.path.isdir(self.publish_log_filename):
            os.remove(self.publish_log_filename)

        with open(self.publish_log_filename, 'ab') as log_file:
            try:
                subprocess.check_call(['git', 'clone', self.clone_url, '--branch', self.publish_branch], stdout=log_file, stderr=subprocess.STDOUT)  # nosec - All subprocess calls use full path
            except subprocess.CalledProcessError as error:
                subprocess.check_call(['git', 'clone', self.clone_url], stdout=log_file, stderr=subprocess.STDOUT)  # nosec - All subprocess calls use full path
                os.chdir(self.clone_dir)
                subprocess.check_call(['git', 'checkout', '-b', self.publish_branch], stdout=log_file, stderr=subprocess.STDOUT)  # nosec - All subprocess calls use full path
                os.chdir('..')
            if not os.path.exists(self.clone_dir):
                raise DocPublishError(f"Repo directory {self.clone_dir} is missing after git clone")

    def get_clone_url(self) -> str:
        """
        Determine the git clone url for the current git repo

        Returns
        -------
        str:
            The git url of the current repo
        """
        self._log_message('\n- Getting the clone url', self.publish_log_filename)
        cwd = os.getcwd()
        os.chdir(self.source_dir)
        try:
            output = subprocess.check_output(['git', 'remote', 'show', 'origin'], stderr=subprocess.DEVNULL)  # nosec - All subprocess calls use full path
        except subprocess.CalledProcessError:  # pragma: no cover
            os.chdir(cwd)
            return ''
        for line in output.decode(errors='ignore').split(os.linesep):
            line = line.strip()
            if line.startswith('Fetch URL:'):
                os.chdir(cwd)
                return ':'.join(line.split(':')[1:]).strip()
        os.chdir(cwd)
        return ''

    def get_clone_dir(self) -> str:
        """
        Return the name of the directory that will be created when git clone runs

        Returns
        -------
        str
            git clone directory
        """
        self._log_message('Determining the clone directory', self.publish_log_filename)
        result = self.get_clone_url().split('/')[-1].removesuffix('.git')
        return result

    def git_add_all(self):
        """
        Run 'git add' on all the files in the current directory.
        """
        self._log_message('\n- Adding all files in the current directory to git', self.publish_log_filename)
        for filename in os.listdir('.'):
            if filename in ['.git']:
                continue
            logger.debug('Adding %s to git', filename)
            self._run_command(['git', 'add', filename], log_filename=self.publish_log_filename)

    def git_commit_documentation(self, message: str = 'Update documentation'):
        """
        Run git commit on all files in the current repo

        Parameters
        ----------
        message: str, optional
            The commit message to use
        """
        self._log_message('\n- Committing git changes', self.publish_log_filename)
        self._run_command(['git', 'commit', '-a', '-m', message], log_filename=self.publish_log_filename)

    def git_push_documentation(self):  # pragma: no cover
        """
        Push the current repository
        """
        self._log_message(f'\n- Pushing the documentation to the {self.publish_branch} branch', self.publish_log_filename)
        self._run_command(['git', 'push', 'origin', self.publish_branch], log_filename=self.publish_log_filename, timeout=self.git_command_timeout)

    def disable_jekyll(self):
        """
        Disable github's jekyll processing for the current repo
        """
        self._log_message('\n- Writing the .nojekyll file to disable github jekyll processing', self.publish_log_filename)
        disable_file = open('.nojekyll', 'w')
        disable_file.close()
        self._run_command(['git', 'add', '.nojekyll'], log_filename=self.publish_log_filename)

    def build_setup(self):
        """
        Run operations to setup the build
        """

    def build_cleanup(self):
        """
        Run operations to cleanup after the build
        """

    def build_documentation(self) -> str:
        """
        Build the documentation with the documentation build tool

        Raises
        ------
        DocBuildError:
            The documentation generate command returned an error

        Returns
        -------
        str:
            The root directory of the generated documenation
        """
        self._log_message(f'\n- Building the {self.name} format documentation', self.build_log_filename)

        self.build_setup()

        cwd = os.getcwd()
        os.chdir(self.source_dir)
        os.makedirs(self.build_dest, exist_ok=True)
        os.makedirs(self.log_dir, exist_ok=True)

        # self.remove_build_log()

        try:
            self._run_command(self.build_command, log_filename=self.build_log_filename)
        finally:
            os.chdir(cwd)

        self.build_cleanup()
        return self.build_dest

    def publish_documentation(self, clear_before_build=True, push=True):  # pragma: no cover
        """
        Build and publsh the documentation

        Raises
        ------
        DocumentationBuildError:
            If the build operation failed

        DocumentationPublishError:
            The publish to the destination failed
        """
        self._log_message(f'\n- Publishing the documentation to github pages', self.publish_log_filename)
        # self.remove_publish_log()

        # Build the documentation
        self.build_documentation()

        if not push:
            return

        # Make sure there is a clone url before trying to publish
        if not self.clone_url:  # pragma: no cover
            raise DocPublishError('Unable to determine a valid git clone url to publish to')

        # Checkout the documentation branch into a temporary directory, add the docs and commit them
        with tempfile.TemporaryDirectory() as tempdir:    # pragma: no cover
            os.chdir(tempdir)
            self.clone_documentation_branch()
            if clear_before_build:
                self.clean_directory(self.clone_dir)
            self.copy_contents(self.build_documentation(), self.clone_dir)

            self._log_message('New repository contents', self.publish_log_filename)
            self._run_command(['ls', '-lR'], log_filename=self.publish_log_filename)

            os.chdir(self.clone_dir)
            self.disable_jekyll()

            self.git_add_all()
            self.git_commit_documentation()
            self.git_push_documentation()

        os.chdir(self.source_dir)


def documentation_plugins(documentation_formats=None):
    """
    Get documentation plugin instances
    """
    for entry_point in importlib.metadata.entry_points(group='screwdrivercd.documentation.plugin'):
        try:
            instance = entry_point.load()()
        except Exception as e:
            logger.warning(f'Documentation plugin {entry_point} failed to load due to a package dependency conflict: {e}')
            raise
        if documentation_formats and instance.name not in documentation_formats:
            continue
        yield instance


def generate_changelog():
    """
    Generate a changelog if the CHANGELOG_FILENAME is set
    """
    changelog_filename = os.environ.get('CHANGELOG_FILENAME', '')
    if not changelog_filename:
        return

    write_changelog(changelog_filename)


def build_documentation(documentation_formats=None):
    """
    Generate documentation using all plugins that can generate documentation
    """
    failed = []
    generate_changelog()
    for documentation_plugin in documentation_plugins(documentation_formats=documentation_formats):
        if documentation_plugin.documentation_is_present:

            print(f'Building {documentation_plugin.name!r} documentation: ', end='', flush=True)
            try:
                documentation_plugin.build_documentation()
                print(colored('Ok', color='green'), flush=True)
            except DocBuildError as error:
                print(colored('Failed', color='red'), flush=True)
                logger.error(f'Building {documentation_plugin.name!r} documentation {colored("Failed", color="red")}')
                if documentation_plugin.build_log_filename and os.path.exists(documentation_plugin.build_log_filename):  # pragma: no cover
                    with open(documentation_plugin.build_log_filename, 'r') as build_log:
                        if logger.level <= logging.DEBUG:
                            logger.error(f'{documentation_plugin.name} build Failed {str(error)}')
                            logger.error(build_log.read())
                        else:
                            print(build_log.read(), flush=True)
                else:  # pragma: no cover
                    logger.debug('No output from the build command was logged')
                failed.append(documentation_plugin.name)

            if documentation_plugin.build_log_filename and os.path.exists(documentation_plugin.build_log_filename):  # pragma: no cover
                with open(documentation_plugin.build_log_filename, 'r') as build_log:
                    logger.debug(f'{documentation_plugin.name} build output')
                    logger.debug(build_log.read())

            if failed:
                raise DocBuildError(f'Documentation build failed for: {" ".join(failed)}')


def publish_documentation(documentation_formats=None, push: bool=True):  # pragma: no cover
    """
    Publish documentation using all plugins that can generate documentation
    """
    clear_before_build = True
    failed = []
    generate_changelog()
    for documentation_plugin in documentation_plugins(documentation_formats=documentation_formats):
        if documentation_plugin.documentation_is_present:
            print(f'Publishing {documentation_plugin.name!r} documentation: ', end='', flush=True)
            try:
                documentation_plugin.publish_documentation(clear_before_build=clear_before_build, push=push)
                print(colored('Ok', color='green'), flush=True)
            except (DocBuildError, DocPublishError) as error:
                print(colored('Failed', color='red'), flush=True)
                if documentation_plugin.build_log_filename and os.path.exists(documentation_plugin.build_log_filename):
                    with open(documentation_plugin.build_log_filename, 'r') as build_log:
                        logger.error(f'{documentation_plugin.name} build Failed {str(error)}')
                        logger.error(build_log.read())
                if documentation_plugin.publish_log_filename and os.path.exists(documentation_plugin.publish_log_filename):
                    with open(documentation_plugin.publish_log_filename, 'r') as publish_log:
                        logger.error(f'{documentation_plugin.name} publish failed {str(error)}')
                        logger.error(publish_log.read())
                error.plugin = documentation_plugin
                failed.append(documentation_plugin.name)
            clear_before_build = False

            if documentation_plugin.build_log_filename and os.path.exists(documentation_plugin.build_log_filename):
                with open(documentation_plugin.build_log_filename, 'r') as build_log:
                    logger.debug(f'{documentation_plugin.name} build output')
                    logger.debug(build_log.read())

            if documentation_plugin.publish_log_filename and os.path.exists(documentation_plugin.publish_log_filename):
                with open(documentation_plugin.publish_log_filename, 'r') as publish_log:
                    logger.debug(f'{documentation_plugin.name} publish output')
                    logger.debug(publish_log.read())

            if failed:
                raise DocPublishError(f'Documentation publish failed for: {" ".join(failed)}')



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""Install apk dependencies"""
import logging
from typing import List

from ..installer import Installer


LOG = logging.getLogger(__name__)


class ApkInstaller(Installer):
    """
    Alpine Package Installer
    """
    install_command: List[str] = ['apk', 'add']
    config_section: str = 'apk'
    install_command_path: List[str] = ['/sbin']



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Type hint validation wrapper for screwdrivercd

This wrapper runs the validation tool.  This wrapper does the following:

- Runs the type checking tool with command line arguments to generate output in the formats specified by TYPE_CHECK_REPORT_FORMAT
- ensures all the logs and reports are stored in the build artifacts before exiting.
- Propagates a success code if the TYPE_CHECK_ENFORCING is set to false.
"""
# The logging_basicConfig has to be run before other imports because some modules we use log output on import
# pylint: disable=wrong-import-order, wrong-import-position
from ..screwdriver.environment import logging_basicConfig, update_job_status
logging_basicConfig(check_prefix='TYPE_CHECK')

import logging
import os
import subprocess  # nosec
import sys

from termcolor import colored
from ..utility import create_artifact_directory, env_bool
from ..utility.package import PackageMetadata, package_srcdir



logger_name = 'validate_type' if __name__ == '__main__' else __name__
logger = logging.getLogger(logger_name)


def validate_with_mypy(report_dir):
    """Run the mypy command directly to do the validation"""

    src_dir = package_srcdir()
    package_name = PackageMetadata().metadata['name']

    # Generate the command line from the environment settings
    command = [sys.executable, '-m', 'mypy']

    # Add report arguments
    reports = os.environ.get('TYPE_CHECK_REPORT_FORMAT', 'txt').split(',')
    for report in [_.strip() for _ in reports]:
        if report in ['junit-xml']:
            command += ['--junit-xml', os.path.join(report_dir, 'mypy.xml')]
            continue
        command += [f'--{report}-report', report_dir]

    # Add extra arguments
    extra_args = os.environ.get('MYPY_ARGS', '')
    if extra_args:
        command += extra_args.split()

    # Add targets
    if src_dir not in ['', '.']:
        command += [src_dir]
    else:
        command += ['-p', package_name]

    print('-' * 90 + '\nRunning:', ' '.join(command) + '\n' + '-' * 90, flush=True)
    rc = 0
    try:
        output = subprocess.check_output(command)  # nosec
    except subprocess.CalledProcessError as error:
        rc = error.returncode
        output = error.output

    for line in output.decode(errors='ignore').split(os.linesep):
        line = line.strip()
        if 'error:' in line:
            print(colored(line, 'red'), flush=True)
        else:
            print(line, flush=True)

    text_report = os.path.join(report_dir, 'index.txt')
    if os.path.exists(text_report):
        with open(text_report) as fh:
            print(fh.read(), flush=True)

    return rc


def validate_type():
    """
    Run a type validator and store the output

    Returns
    -------
    int:
        Exit returncode from the type validation command or 0 if TYPE_CHECK_ENFORCING env variable is false
    """
    logging_basicConfig()

    # Set the status message
    # update_job_status(status='SUCCESS', message='Checking type annotations')

    # Make sure the report directory exists
    artifacts_dir = os.environ.get('SD_ARTIFACTS_DIR', '')
    report_dir = os.path.join(artifacts_dir, 'reports/type_validation')
    create_artifact_directory(report_dir)

    rc = validate_with_mypy(report_dir=report_dir)

    if rc > 0 and env_bool('TYPE_CHECK_ENFORCING'):
        print(colored('ERROR: Type check failed', 'red'), file=sys.stderr, flush=True)
        # update_job_status(status='FAILURE', message='Type annotation check failed')
        return rc
    if rc == 0:
        print(colored('OK: Type validation sucessful', 'green'), flush=True)
        # update_job_status(status='SUCCESS', message='Type annotation check passed')
    else:
        print(colored('WARNING: Type check failed, enforcement is disabled, so not failing check', 'yellow'))
        update_job_status(status='SUCCESS', message='Type check, failed but is not enforcing')
        return 0
    return rc


def main():
    """
    Type check runner utility command line entry point

    Returns
    -------
    int:
        Returncode from running the check
    """
    return validate_type()


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
screwdrivercd.documentation module entrypoint
"""
import sys
from .cli import main


sys.exit(main())



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
screwdrivercd.documentation module command exceptions
"""
class DocBuildError(Exception):
    """Generic Documentation Generation Error"""
    plugin: str=''


class DocPublishError(Exception):
    """Documentation Publication Error"""
    plugin: str=''



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Python packaging publish command line wrapper
"""
from ..screwdriver.environment import logging_basicConfig, update_job_status
logging_basicConfig(check_prefix='PUBLISH_PYTHON')

import logging
import os
import subprocess  # nosec
import sys
import time
from datetime import datetime, timedelta
from typing import Dict, List, Optional, Set

from packaging.utils import canonicalize_name
import requests

from ..utility.environment import env_bool, interpreter_bin_command, standard_directories
from ..utility.package import setup_query


logger = logging.getLogger(__name__)


def package_exists(package: str, package_filename: str, endpoint: str='https://pypi.org/simple') -> bool:
    url = f'{endpoint}/{canonicalize_name(package)}/'

    # Stream the index
    req = requests.get(url, stream=True)
    for line in req.iter_lines():
        line = line.decode(errors='ignore')
        if package_filename in line:
            return True
    return False


def poll_until_available(package: str, packages: Set[str], endpoint: str='https://pypi.org/simple', timeout=300, poll_interval=15) -> Set[str]:
    """
    Poll the packaging repository until package files show up in the index

    Parameters
    ----------
    package: str
        The package the package files are published for

    packages: set of str
        Packages to poll for

    endpoint: str, optional
        The packaging repository endpoint to poll, default=https://pypi.org/simple

    timeout: int, optional
        Timeout in seconds, default=300

    poll_interval: int, optional
        Frequency to poll the package repository

    Returns
    ------
    set:
        Set of packages that failed to publish, empty set if no packages failed
    """
    start = datetime.utcnow()
    end = start + timedelta(seconds=timeout)
    completed: Set[str] = set()

    delay = 1
    while datetime.utcnow() < end:
        for filename in packages - completed:
            if package_exists(package, filename, endpoint=endpoint):
                completed.add(filename)

        elapsed = datetime.utcnow() - start
        not_published = packages - completed
        print(f'\t{elapsed}: Published: {list(completed)}, Waiting for: {list(not_published)}', flush=True)
        
        if completed == packages:
            return set()

        if delay < poll_interval:
            delay = delay * 2
        if delay > poll_interval:  # pragma: no cover
            delay = poll_interval
        time.sleep(delay)
    return packages - completed


def main(twine_command: str='') -> int:
    if not twine_command:  # pragma: no cover
        twine_command = interpreter_bin_command('twine')

    directories = standard_directories(command='publish_python')
    user = os.environ.get('PYPI_USER', None)
    password = os.environ.get('PYPI_PASSWORD', None)
    publish_timeout = int(os.environ.get('PUBLISH_PYTHON_TIMEOUT', os.environ.get('ARTIFACTORY_TIMEOUT', "300")))

    twine_repository_url = os.environ.get('TWINE_REPOSITORY_URL', 'https://upload.pypi.org/legacy/')

    if not os.path.exists(directories['packages']):
        print(f'Package directory {directories["packages"]!r} does not exist, no packages to publish', flush=True)
        return 0

    if not env_bool('PUBLISH_PYTHON', True):
        print('Publish is disabled, skipping publish operation', flush=True)
        return 0

    if not env_bool('PUBLISH', True):
        print('Publish is disabled, skipping publish operation', flush=True)
        return 0

    bad_cred_rc = 0
    if env_bool('PUBLISH_PYTHON_FAIL_MISSING_CRED', False):
        bad_cred_rc = 1

    if 'test.pypi.org' in twine_repository_url:
        print('Using test.pypi.org endpoint, getting user from TEST_PYPI_USER secret and password from TEST_PYPI_PASSWORD secret', flush=True)
        user = os.environ.get('TEST_PYPI_USER', None)
        password = os.environ.get('TEST_PYPI_PASSWORD', None)

    if not user:
        print('Unable to publish to PYPI, user secret is not set', flush=True)
        return bad_cred_rc

    if not password:
        print('Unable to publish to PYPI, password secret is not set', flush=True)
        return bad_cred_rc

    if not os.path.exists(twine_command):  # pragma: no cover
        print('The twine command is missing', flush=True)
        return 1

    print(f'\n# Publishing to {twine_repository_url} as user {user}', flush=True)

    package_name = setup_query('--name')
    published = []
    failed = []
    for filename in os.listdir(directories['packages']):
        print(f'\tUploading {filename}', flush=True)
        command = [twine_command, 'upload', '--verbose', os.path.join(directories['packages'], filename)]

        os.environ['TWINE_USERNAME'] = user
        os.environ['TWINE_PASSWORD'] = password
        os.environ['TWINE_REPOSITORY_URL'] = twine_repository_url
       
        print(f'\tRunning: {" ".join(command)}')
        log_filename = os.path.join(directories['logs'], f'twine_{filename}.log')
        with open(log_filename, 'ab') as output_handle:
            try:
                subprocess.check_call(command, stdout=output_handle, stderr=subprocess.STDOUT)  # nosec
                published.append(filename)
            except subprocess.CalledProcessError as error:
                print(f'\tUpload of package file {filename!r} failed, returncode {error.returncode}', flush=True)
                output_handle.write(f'Upload of package file {filename!r} failed, returncode {error.returncode}'.encode(errors='ignore'))
                failed.append(filename)
                if error.stdout:
                    output_handle.write(error.stdout)
                if error.stderr:
                    output_handle.write(error.stderr)

    endpoint = twine_repository_url.rstrip('/')
    if endpoint.endswith('/legacy'):
        endpoint = endpoint[:-7] + '/simple'
    if endpoint.startswith('https://upload.'):
        endpoint = f'https://{endpoint[15:]}'

    publish_failed: List[str] = []
    if publish_timeout:
        print(f'\n# Polling {endpoint}/{package_name}/ for updated pacakge')
        publish_failed = list(poll_until_available(package_name, set(published), endpoint=endpoint, timeout=publish_timeout))

    if publish_failed:  # pragma: no cover
        failed += publish_failed

    if failed:
        update_job_status(status='FAILURE', message=f'{len(failed)} packages failed')
        return len(failed)

    return 0


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Command execution utilities
"""
import os
import subprocess  # nosec
from typing import List


def run_and_log_output(command: List[str], logfile: str, print_errors: bool=True):
    """
    Run a command with subprocess.check_output and send the output to the console and the specified logfile

    Parameters
    ----------
    command: list of str
        The parsed command to execute

    logfile: str
        The full path to the logfile to hold the output

    Raises
    ------
    subprocess.CalledProcessError - The command failed
    """
    log_dir = os.path.dirname(logfile)
    if log_dir:
        os.makedirs(log_dir, exist_ok=True)

    with open(logfile, 'wb') as fh:
        print(f'Running command: {" ".join(command)}', flush=True)
        try:
            output = subprocess.check_output(command, stderr=subprocess.STDOUT)  # nosec
            fh.write(output)
        except subprocess.CalledProcessError as error:
            print(f'Command {command!r} failed', flush=True)
            fh.write(error.stdout)
            if print_errors and error.stdout:  # pragma: no cover
                print(error.stdout.decode(errors='ignore'))
            raise error



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""Install rpm dependencies"""
import logging
import os
import shutil
import subprocess  # nosec - All subprocess calls use full path
from typing import List

from termcolor import colored

from ..installer import Installer


LOG = logging.getLogger(__name__)


class AptInstaller(Installer):
    """
    Debian/Ubuntu Package Installer
    """
    install_command: List[str] = ['apt-get', 'install', '-y']
    install_repo_command: List[str] = ['add-apt-repository']
    config_section: str = 'apt-get'
    install_command_path: List[str] = ['/usr/bin']
    supports_repositories: bool = True
    _repo_tool_install_failed: bool = False
    _updated_index: bool = False

    def update_index(self):  # pragma: no cover - Function is OS specific
        """
        Method to update the package index
        """
        if self._updated_index:
            return
        subprocess.check_call([self.install_command[0], 'update'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)  # nosec - All subprocess calls use full path
        self._updated_index = True

    def install_repo_tool(self):  # pragma: no cover - Function is OS specific
        """
        Install the repo install tool
        """
        if self.install_repo_command[0].startswith('/') or self._repo_tool_install_failed:
            return

        install_repo_command = shutil.which(self.install_repo_command[0])
        if install_repo_command:
            self.install_repo_command[0] = install_repo_command

        if not os.path.exists(self.install_repo_command[0]):
            LOG.debug(f'Tool to add repos, {self.install_repo_command[0]!r} is missing, trying to install it')
            subprocess.check_call([self.install_command[0], 'install', '-y', 'software-properties-common'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)  # nosec - All subprocess calls use full path
            install_repo_command = shutil.which(self.install_repo_command[0])
            if install_repo_command:
                self.install_repo_command[0] = install_repo_command
            else:
                self._repo_tool_install_failed = True
                raise FileNotFoundError(f'Could not find the {self.install_repo_command[0]} utility')

    def add_repo(self, repo_name, repo_url):  # pragma: no cover - Function is OS specific
        """
        Add apt repos specified in the configuration
        """
        try:
            self.install_repo_tool()
        except FileNotFoundError:
            message = f'The {self.install_repo_command[0]} utility is missing and cannot be installed, cannot add the {repo_name} repository'
            if self.print_output:
                print(colored(message, 'red'), flush=True)
            else:
                LOG.error(message)
            return

        if self.print_output:
            print(f'Enabling {repo_name!r} repo: ', end='', flush=True)
        try:
            subprocess.check_call(self.install_repo_command + [repo_url], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)  # nosec - All subprocess calls use full path
            if self.print_output:
                print(colored('Ok', 'green'), flush=True)
        except subprocess.CalledProcessError:
            if self.print_output:
                print(colored('Failed', 'red'), flush=True)
            else:
                LOG.error(f'Adding {repo_name!r} repo failed')



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Source code repository management utilities
"""
import os
import subprocess  # nosec
import sys

from tempfile import TemporaryDirectory

from ..changelog.generate import changelog_contents
from ..utility.environment import env_bool
from ..version.version_types import Version


def create_release_tag(version: str, git_command: str='git', message: str=''):

    if not message:
        try:
            message = subprocess.check_output([git_command, 'log', '-n', '1']).decode(errors='ignore').split('\n')[4].strip()  # nosec
        except (subprocess.CalledProcessError, KeyError):
            message = f'New release {version}'

    with TemporaryDirectory() as tdir:
        tfilename = os.path.join(tdir, 'message')
        with open(tfilename, 'w') as fh:
            fh.write(message)
        command = [git_command, 'tag', '-a', f'v{version}', '-F', tfilename, '-f']
        subprocess.call(command)  # nosec


def push_release_tag(git_command: str='git', timeout=60):
    try:
        subprocess.check_call([git_command, 'push', '-f', '--tags'], timeout=timeout)  # nosec
    except subprocess.TimeoutExpired:
        print('\tTimeout occurred pushing tags to the remote', flush=True)
        return
    except subprocess.CalledProcessError:
        print('\tPush of git tag failed', flush=True)
    return


def main(meta_command: str='meta') -> int:

    if not env_bool('PUBLISH', True):
        print('Publish is disabled, skipping tag operation', flush=True)
        return 0

    if not env_bool('PACKAGE_TAG', True):
        print('Tagging is disabled for this job')
        return 0

    # If ssh-agent isn't already configured, use the GIT_DEPLOY_KEY
    if not os.environ.get('SSH_AUTH_SOCK', ''):  # pragma: no cover
        if not os.environ.get('GIT_DEPLOY_KEY', ''):
            print('Git deployment key is not present, cannot commit tags to the git repo')
            return 0

    version = Version(meta_command=meta_command).meta_version
    if not version:  # pragma: no cover
        print('No release version in metadata', flush=True)
        return 0

    print('# Creating tag', flush=True)
    create_release_tag(version=version)
    changelog = changelog_contents(f'v{version}').replace('\n#', '\n-')
    create_release_tag(version=version, message=changelog)

    print('\n# Pushing the tags', flush=True)
    push_release_tag()
    return 0


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
import os
from pathlib import Path
from unittest import skip

import screwdrivercd.documentation.exceptions
import screwdrivercd.documentation.plugin
import screwdrivercd.documentation.mkdocs.plugin
import screwdrivercd.documentation.sphinx.plugin

from . import ScrewdriverTestCase


# Simple sphinx project directory
sphinx_project =  {
    'doc/source/conf.py': b"""# -*- coding: utf-8 -*-
import os
extensions = []
templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'
project = 'foo'
copyright = ''
version = '0.0.0'
release = version
""",
    'doc/source/index.rst': b"""Test file"""
}


mkdocs_project_config = {
    'mkdocs.yml': b'site_name: test\nsite_url: https://foo.bar.com/\nstrict: true\nnav:\n    - foo: foo.md\n',
    'docs/index.md': b"# Test file\n",
    'docs/foo.md': b"# Test file\n",
}

mkdocs_project_config_fail = {
    'mkdocs.yml': b'site_name: test\nsite_url: https://foo.bar.com/\nstrict: true\nnav:\n    - foo: foo2.md\n',
    'docs/index.md': b"# Test file\n",
    'docs/foo.md': b"# Test file\n",
}


class BuildDocsTestCase(ScrewdriverTestCase):
    def _init_test_repo(self):
        os.system('git init')
        os.system('git config user.email "foo@bar.com"')
        os.system('git config user.name "foo the bar"')
        os.system('git remote add origin https://github.com/yahoo/python-screwdrivercd.git')

    def test__doc_build(self):
        self._init_test_repo()
        self.write_config_files(mkdocs_project_config)
        screwdrivercd.documentation.plugin.build_documentation()

    def test__doc_build_fail(self):
        self._init_test_repo()
        self.write_config_files(mkdocs_project_config_fail)
        with self.assertRaises(screwdrivercd.documentation.exceptions.DocBuildError):
            screwdrivercd.documentation.plugin.build_documentation()

    def test__doc_publish(self):
        self._init_test_repo()
        self.write_config_files(mkdocs_project_config)
        screwdrivercd.documentation.plugin.publish_documentation(push=False)

    def test__doc_publish_fail(self):
        self._init_test_repo()
        self.write_config_files(mkdocs_project_config_fail)
        with self.assertRaises(screwdrivercd.documentation.exceptions.DocPublishError):
            screwdrivercd.documentation.plugin.publish_documentation(push=False)


class PluginsTestCase(ScrewdriverTestCase):

    def test__documentation_plugins__present(self):
        names = [_.name for _ in screwdrivercd.documentation.plugin.documentation_plugins()]
        self.assertIn('mkdocs', names)
        self.assertIn('mkdocs_venv', names)
        self.assertIn('sphinx', names)


class DocumentationPluginTestCase(ScrewdriverTestCase):
    plugin_class = screwdrivercd.documentation.plugin.DocumentationPlugin

    def _create_test_repo_contents(self):
        pass

    def _init_test_repo(self):
        os.system('git init')
        os.system('git config user.email "foo@bar.com"')
        os.system('git config user.name "foo the bar"')
        os.system('git remote add origin https://github.com/yahoo/python-screwdrivercd.git')

    def test__get_clone_url(self):
        self._init_test_repo()
        p = screwdrivercd.documentation.plugin.DocumentationPlugin()
        result = p.get_clone_url()
        self.assertEqual(result, 'https://github.com/yahoo/python-screwdrivercd.git')

    def test__get_clone_dir(self):
        self._init_test_repo()
        p = screwdrivercd.documentation.plugin.DocumentationPlugin()
        result = p.get_clone_dir()
        self.assertEqual(result, 'python-screwdrivercd')

    def test__clone_url(self):
        self._init_test_repo()
        p = screwdrivercd.documentation.plugin.DocumentationPlugin()
        result = p.clone_url
        self.assertEqual(result, 'https://github.com/yahoo/python-screwdrivercd.git')

    def test__clone_dir(self):
        self._init_test_repo()
        p =  screwdrivercd.documentation.plugin.DocumentationPlugin()
        result = p.clone_dir
        self.assertEqual(result, 'python-screwdrivercd')

    def test__git_add_all(self):
        self._init_test_repo()
        p = screwdrivercd.documentation.plugin.DocumentationPlugin()
        Path('testfile.md').touch()
        p.git_add_all()

    def test__disable_jekyll(self):
        self._init_test_repo()
        p = screwdrivercd.documentation.plugin.DocumentationPlugin()
        p.disable_jekyll()

    def test_documentation_base_plugin_is_present(self):
        p = screwdrivercd.documentation.plugin.DocumentationPlugin()
        self.assertFalse(p.documentation_is_present)

    def test_documentation_base__log_message__default(self):
        p = self.plugin_class()
        p.build_log_filename = 'foo.log'
        p._log_message('foo')
        self.assertTrue(os.path.exists('foo.log'))
        with open('foo.log') as log:
            self.assertEqual('foo\n', log.read())

    def test_documentation_base__log_message__log_filename(self):
        p = self.plugin_class()
        p.build_log_filename = 'foo2.log'
        p._log_message('foo', log_filename='foo.log')
        self.assertTrue(os.path.exists('foo.log'))
        with open('foo.log') as log:
            self.assertEqual('foo\n', log.read())

    def test_documentation_base__log_message__pass_end(self):
        p = self.plugin_class()
        p._log_message('foo', log_filename='logs/foo.log', end='\r\n')
        self.assertTrue(os.path.exists('logs/foo.log'))
        with open('logs/foo.log', 'rb') as log:
            self.assertEqual(b'foo\r\n', log.read())

    def test_documentation_base__log_message_pass_log_filename(self):
        p = self.plugin_class()
        p.build_log_filename = 'foo.log'
        p._log_message('foo', log_filename='foo2.log')
        self.assertTrue(os.path.exists('foo2.log'))
        with open('foo2.log') as log:
            self.assertEqual('foo\n', log.read())

    def test_documentation__base__run_command(self):
        p = self.plugin_class()
        p.build_log_filename = 'foo.log'
        p._run_command(['echo', 'hello'], log_filename=p.build_log_filename)
        self.assertTrue(os.path.exists('foo.log'))
        with open('foo.log') as log:
            self.assertEqual('hello\n', log.read())

    def test_documentation__base__run_command__error(self):
        p = self.plugin_class()
        p.build_log_filename = 'foo.log'
        with self.assertRaises(screwdrivercd.documentation.exceptions.DocBuildError):
            p._run_command(['./exit', '1'], log_filename=p.build_log_filename)

    def test__documentation__build(self):
        self._create_test_repo_contents()
        self.write_config_files(mkdocs_project_config)
        p = self.plugin_class()
        p.build_documentation()
        self.assertTrue(os.path.exists(f'{p.log_dir}/{p.name}.build.log'))

    def test__remove_build_log(self):
        p = self.plugin_class()
        p.build_log_filename = os.path.join(self.tempdir.name, 'build.log')
        Path(p.build_log_filename).touch()
        self.assertTrue(os.path.exists(p.build_log_filename))
        p.remove_build_log()
        self.assertFalse(os.path.exists(p.build_log_filename))

    def test__remove_publish_log(self):
        p = self.plugin_class()
        p.publish_log_filename = os.path.join(self.tempdir.name, 'publish.log')
        Path(p.publish_log_filename).touch()
        self.assertTrue(os.path.exists(p.publish_log_filename))
        p.remove_publish_log()
        self.assertFalse(os.path.exists(p.publish_log_filename))

    def test__clean_directory(self):
        p = self.plugin_class()
        testdir = Path(self.tempdir.name) / 'testdir'
        testdir.mkdir(exist_ok=True)
        testfile = testdir / 'testfile'
        testfile.touch()
        p.clean_directory(str(testdir))
        self.assertFalse(testfile.exists())

    def test__copy_contents(self):
        dir1 = Path(self.tempdir.name) / 'dir1'
        dir2 = Path(self.tempdir.name) / 'dir2'
        srctestfile = dir1 / 'testfile'
        srctestdotfile = dir1 / '.testfile'
        dsttestfile = dir2 / 'testfile'
        dsttestdotfile = dir2 / '.testfile'

        dir1.mkdir(exist_ok=True)
        dir2.mkdir(exist_ok=True)
        srctestfile.touch()
        srctestdotfile.touch()

        p = self.plugin_class()
        p.copy_contents(str(dir1), str(dir2))

        self.assertTrue(dsttestfile.exists())
        self.assertTrue(dsttestdotfile.exists())

    def test__documentation__publish(self):
        instance = self.plugin_class()
        instance.publish_documentation(push=False)

    def test_git_command_timeout_valid(self):
        os.environ['DOCUMENTATION_GIT_TIMEOUT'] = '600'
        p = self.plugin_class()
        self.assertEqual(p.git_command_timeout, 600)

    def test_git_command_timeout_invalid_type(self):
        os.environ['DOCUMENTATION_GIT_TIMEOUT'] = 'invalid'
        p = self.plugin_class()
        self.assertEqual(p.git_command_timeout, 300)  # Default value

    def test_git_command_timeout_not_set(self):
        if 'DOCUMENTATION_GIT_TIMEOUT' in os.environ:
            del os.environ['DOCUMENTATION_GIT_TIMEOUT']
        p = self.plugin_class()
        self.assertEqual(p.git_command_timeout, 300)  # Default value


class SphinxDocumentationPluginTestCase(DocumentationPluginTestCase):
    plugin_class = screwdrivercd.documentation.sphinx.plugin.SphinxDocumentationPlugin

    def _create_test_repo_contents(self):
        pass

    def test__documentation__build(self):
        pass

    def test__documentation__publish(self):
        self.write_config_files(sphinx_project)
        super().test__documentation__publish()


class MkdocsDocumentationPluginTestCase(DocumentationPluginTestCase):
    plugin_class = screwdrivercd.documentation.mkdocs.plugin.MkDocsDocumentationPlugin

    def _create_test_repo_contents(self):
        pass

    def test__documentation__publish(self):
        self._init_test_repo()
        self.write_config_files(mkdocs_project_config)
        super().test__documentation__publish()


class MkdocsDocumentationVenvPluginTestCase(DocumentationPluginTestCase):
    plugin_class = screwdrivercd.documentation.mkdocs.plugin.MkDocsDocumentationVenvPlugin

    def _create_test_repo_contents(self):
        pass

    def test__documentation__publish(self):
        self._init_test_repo()
        self.write_config_files(mkdocs_project_config)
        super().test__documentation__publish()

    def test__documentation__publish__unchanged(self):
        self._init_test_repo()
        self.write_config_files(mkdocs_project_config)
        p = screwdrivercd.documentation.mkdocs.plugin.MkDocsDocumentationVenvPlugin()
        p.build_setup()
        p.build_documentation()
        p.build_cleanup()
        p.publish_documentation(push=False)
        os.system(f'git checkout {p.publish_branch}')
        p.build_setup()
        p.build_documentation()
        p.build_cleanup()
        p.publish_documentation(push=False)



#!/usr/bin/env python
# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Code to generate a changelog for a git repository
"""
import logging
import os
import subprocess  # nosec
import sys
from datetime import datetime
from pathlib import Path
from typing import Dict, List

from ..utility.environment import env_bool
from ..utility.package import setup_query
from ..utility.run import run_and_log_output


LOG = logging.getLogger(__name__)
CHANGE_TYPES = dict(
    feature='Features',
    bugfix='Bugfixes',
    doc='Improved Documentation',
    removal='Removed',
    misc='Misc Changes'
)


def git_fetch_tags():
    artifacts_dir = os.environ.get('SD_ARTIFACTS_DIR', '')
    log_filename = os.path.join(artifacts_dir, 'logs/changelog/fetch_tags.log')

    command = ['git', 'fetch', '--tags']
    run_and_log_output(command, logfile=log_filename)


def git_tag_dates() -> Dict[str, str]:
    tags = {}

    command = ['git', 'log', '--date-order', '--tags', '--simplify-by-decoration', '--pretty=format:%ct|%d']
    with subprocess.Popen(command, stdout=subprocess.PIPE) as tag_command:  # nosec
        for line in tag_command.stdout.readlines():
            date, all_tags = line.decode(errors='ignore').strip().split('|')
            all_tags = all_tags.strip().lstrip('(').rstrip(')')

            for tag in all_tags.split(','):
                tag = tag.strip()
                if tag.startswith('tag: '):
                    tag = tag[5:]
                tags[tag] = date
    return tags


def create_first_commit_tag_if_missing() -> None:
    if 'first_commit' in git_tag_dates().keys():
        return

    first_commit_hash = subprocess.check_output(['git', 'rev-list', '--max-parents=0', 'HEAD'], stderr=subprocess.DEVNULL).decode(errors='ignore').strip()  # nosec
    if first_commit_hash:
        try:
            output = subprocess.check_output(['git', 'tag', 'first_commit', first_commit_hash])  # nosec
        except subprocess.CalledProcessError:  # pragma no cover
            pass  # Tag already exists

        
def changed_files(commit1: str, commit2: str, changelog_dir: str='changelog.d') -> List[Path]:
    changed = []

    with subprocess.Popen(['git', 'diff', f'{commit1}..{commit2}', changelog_dir], stdout=subprocess.PIPE) as diff_command:  # nosec
        for binline in diff_command.stdout.readlines():
            line = binline.decode(errors='ignore').strip()
            if line.startswith('+++'):
                line = line.lstrip('+++ b/')
                filepath = Path(line)
                if filepath.parts[0] != changelog_dir:  # pragma: no cover
                    continue
                changed_file = Path(*filepath.parts)
                changed.append(changed_file)
    return changed


def release_changes(changelog_dir: str, only_versions: bool=True, only_stable: bool=False) -> Dict[str, Dict[str, Dict[str, str]]]:
    create_first_commit_tag_if_missing()
    tags = git_tag_dates()

    previous_commit = commit = 'first_commit'
    commits = list(tags.keys())
    commits.reverse()

    changes: Dict[str, Dict[str, Dict[str, str]]] = {}
    for commit in commits:
        if only_versions:
            if not commit.startswith('v'):
                if commit != 'first_commit':
                    continue
        if only_stable and commit.startswith('v'):
            try:
                int(commit.split('.')[-1])
            except ValueError:
                # Last part of the version has a non-numeric character
                # so it is a pre-release
                continue

        changed = changed_files(previous_commit, commit, changelog_dir=changelog_dir)
        if not changed:
            previous_commit = commit
            continue

        changes[commit] = {}
        for change in changed:
            filename = change.parts[-1]
            if filename in ['README.md', 'FOOTER.md', 'HEADER.md']:
                continue
            split_filename = str(filename).split('.')
            changeid = split_filename[0]
            change_type = split_filename[1]
            if change_type not in CHANGE_TYPES:
                LOG.warning(f'Invalid change type {change_type}')
                previous_commit = commit
                continue
            if change_type not in changes[commit].keys():  # pragma: no cover
                changes[commit][change_type] = {}
            changes[commit][change_type][changeid] = change.read_text().rstrip()

        previous_commit = commit
    return changes


def changelog_contents(changelog_releases: str='') -> str:
    """
    Generate the changelog and return the contents as a string

    Parameters
    ----------
    changelog_releases: str, optional
        Comma separated list of releases to include in the changelog.  If
        not provided, will use all the releases.

    Returns
    -------
    str:
        The generated changelog in Markdown format.
    """
    if not changelog_releases:
        changelog_releases = os.environ.get('CHANGELOG_RELEASES', 'all')

    only_versions = bool(env_bool('CHANGELOG_ONLY_VERSION_TAGS', True))
    only_stable = bool(env_bool('CHANGELOG_ONLY_STABLE_RELEASES', False))
    changelog_dir = os.environ.get('CHANGELOG_DIR', 'changelog.d')

    header_filename = os.path.join(changelog_dir, 'HEADER.md')
    footer_filename = os.path.join(changelog_dir, 'FOOTER.md')

    changelog_name = os.environ.get('CHANGELOG_NAME', '')
    if not changelog_name and os.path.exists('setup.py'):
        try:
            changelog_name = setup_query('--name')
        except subprocess.CalledProcessError:  # pragma: no cover
            changelog_name = ''
    release_dates = git_tag_dates()

    output = ''
    footer = ''
    header = ''
    release_changelog = release_changes(changelog_dir, only_versions=only_versions, only_stable=only_stable)
    if changelog_releases == 'all':
        if os.path.exists(header_filename):
            with open(header_filename) as fh:
                header = fh.read()
        if os.path.exists(footer_filename):
            with open(footer_filename) as fh:
                footer = fh.read()
    else:
        selected_releases = set([_.strip() for _ in changelog_releases.split(',')])
        available_releases = set(list(release_changelog.keys()))

        new_release_changelog = {}
        for release in available_releases.intersection(selected_releases):
            new_release_changelog[release] = release_changelog[release]

        release_changelog = new_release_changelog

    if header:
        output += header + os.linesep

    releases = list(release_changelog.keys())
    releases.reverse()

    for release in releases:
        changes = release_changelog[release]
        if not changes or release in ['first_commit', 'last_commit']:  # pragma: no cover
            continue
        date = datetime.fromtimestamp(int(release_dates[release]))
        if len(releases) > 1:
            output += f'{os.linesep}---{os.linesep}'
        if changelog_name:
            output += f'## {changelog_name} {release} ({date:%Y-%m-%d}){os.linesep}'
        else:  # pragma no cover
            output += f'## {release} ({date:%Y-%m-%d}){os.linesep}'
        for change_type, change_desc in CHANGE_TYPES.items():
            if change_type not in changes.keys():
                continue

            output += f'### {change_desc}{os.linesep}'
            for changeid, change_text in changes[change_type].items():
                output += f'- {change_text}{os.linesep}'
        output += f'{os.linesep}'
    if footer:
        output += os.linesep + footer

    return output


def write_changelog(filename, changelog_releases: str=''):
    report_dir = os.path.dirname(filename)
    os.makedirs(report_dir, exist_ok=True)
    with open(filename, 'w') as report_handle:
        report_handle.write(changelog_contents(changelog_releases=changelog_releases))


def main():
    artifacts_dir = os.environ.get('SD_ARTIFACTS_DIR', 'artifacts')
    report_dir = os.path.join(artifacts_dir, 'reports/changelog')
    report_filename = os.environ.get('CHANGELOG_FILENAME', os.path.join(report_dir, 'changelog.md'))

    git_fetch_tags()
    write_changelog(report_filename)
    return 0


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Documentation Utility functions
"""
import logging
import os
import shutil


logger = logging.getLogger(__name__)  # pylint: disable=C0103


def clean_directory(directory_name):
    """
    Remove all files and folders from a directory that do not begin with '.'

    Parameters
    ----------

    directory_name : str
        Directory to be cleaned up.
    """
    for filename in os.listdir(directory_name):
        if filename.startswith('.'):  # pragma: no cover
            continue
        full_filename = os.path.join(directory_name, filename)
        logger.debug('Removing file: %s', full_filename)
        if os.path.isdir(full_filename):
            shutil.rmtree(full_filename)
        else:
            os.remove(full_filename)


def copy_contents(src, dest, skip_dotfiles=False):
    """
    Copy the contents of the src directory to the dest directory

    Parameters
    ----------

    src : str
        Source directory

    dest : str
        Destination directory

    skip_dotfiles : bool
        Don't copy dotfiles if True.
    """
    logger.debug('Copying: %s -> %s', src, dest)
    cwd = os.getcwd()
    dest = os.path.abspath(dest)
    os.chdir(src)

    for dirname, subdirlist, filelist in os.walk('.'):  # pylint: disable=W0612
        logger.debug('Found directory: %s', dirname)
        if dirname.startswith('./'):  # pragma: no cover
            dirname = dirname[2:]
        destdir = os.path.join(dest, dirname)
        logger.debug('Dest directory: %s', destdir)
        os.makedirs(destdir, exist_ok=True)
        for fname in filelist:
            if skip_dotfiles and fname.startswith('.'):
                continue
            srcfile = os.path.join(dirname, fname)
            destfile = os.path.join(destdir, fname)
            logger.debug('\tcopying file: %s -> %s', srcfile, destfile)
            shutil.copyfile(srcfile, destfile)
    os.chdir(cwd)



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
screwdrivercd.documentation.sphinx module
"""
__all__ = ['plugin']



#!/usr/bin/env python
# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
import json
import logging
logging.basicConfig(level=logging.DEBUG)
import unittest
from screwdrivercd.installdeps.config import Configuration
from screwdrivercd.utility.contextmanagers import InTemporaryDirectory


TEST_CONFIG = '''[build-system]
requires = ["setuptools", "wheel"]  # PEP 508 specifications.

[tool.sdv4_installdeps]
    install = ['apk', 'apt-get', 'yinst', 'yum', 'pip3']

    [tool.sdv4_installdeps.apk]
        deps = [
            'python3',
            'mysql-client'
        ]

    [tool.sdv4_installdeps.apt-get]
        deps = [
            'python3',
            'mysql-client'
        ]
        repos = {}

    [tool.sdv4_installdeps.yum]
        repos.verizon_python_rpms = "https://edge.artifactory.yahoo.com:4443/artifactory/python_rpms/python_rpms.repo"
        deps = [
            'yahoo_python36;distro_version<"7.5',
            'yahoo_python37;distro_version>="7.5"',
            'mysql;distro_version<"7"',
            'mariadb;distro_version>="7"'
        ]

    [tool.sdv4_installdeps.yinst]
        deps = [
            'python36',
            'dist_utils'
        ]
        deps_stable = []
        deps_current = []
        deps_test = []
        deps_quarantine = []

    [tool.sdv4_installdeps.pip3]
        bin_dir = ''
        deps = []
        repos = {}
'''

class TestConfig(unittest.TestCase):
    def setUp(self):
        super(TestConfig, self).setUp()

    def test__configuration__defaults__no_config(self):
        with InTemporaryDirectory():
            result = Configuration()
            self.assertListEqual(result.configuration['apk']['deps'], [])
            self.assertListEqual(result.configuration['apt-get']['deps'], [])
            self.assertListEqual(result.configuration['install'], ['apk', 'apt-get', 'yinst', 'yum', 'pip3'])
            self.assertListEqual(result.configuration['yinst']['deps'], [])
            self.assertListEqual(result.configuration['yum']['deps'], [])
            self.assertListEqual(result.configuration['pip3']['deps'], [])

    def test__configuration__no_tool_configs(self):
        with InTemporaryDirectory():
            with open('pyproject.toml', 'w') as file_handle:
                file_handle.write('[build-system]\nrequires = ["setuptools", "wheel"]  # PEP 508 specifications.')
            result = Configuration('pyproject.toml')
            self.assertListEqual(result.configuration['apk']['deps'], [])
            self.assertListEqual(result.configuration['apt-get']['deps'], [])
            self.assertListEqual(result.configuration['install'], ['apk', 'apt-get', 'yinst', 'yum', 'pip3'])
            self.assertListEqual(result.configuration['yinst']['deps'], [])
            self.assertListEqual(result.configuration['yum']['deps'], [])
            self.assertListEqual(result.configuration['pip3']['deps'], [])

    def test__configuration__invalid_filename(self):
        with InTemporaryDirectory():
            with open('pyproject.toml', 'w') as file_handle:
                file_handle.write('[build-system]\nrequires = ["setuptools", "wheel"]  # PEP 508 specifications.')
            result = Configuration('pyprojectt.toml')
            self.assertListEqual(result.configuration['apk']['deps'], [])
            self.assertListEqual(result.configuration['apt-get']['deps'], [])
            self.assertListEqual(result.configuration['install'], ['apk', 'apt-get', 'yinst', 'yum', 'pip3'])
            self.assertListEqual(result.configuration['yinst']['deps'], [])
            self.assertListEqual(result.configuration['yum']['deps'], [])
            self.assertListEqual(result.configuration['pip3']['deps'], [])

    def test__configuration__no_sdv4_installdeps_configs(self):
        with InTemporaryDirectory():
            with open('pyproject.toml', 'w') as file_handle:
                file_handle.write('[build-system]\nrequires = ["setuptools", "wheel"]  # PEP 508 specifications.\n[tool.foo]\ninstall = ["apk", "apt-get", "yinst", "yum", "pip3"]')
            result = Configuration()
            self.assertListEqual(result.configuration['apk']['deps'], [])
            self.assertListEqual(result.configuration['apt-get']['deps'], [])
            self.assertListEqual(result.configuration['install'], ['apk', 'apt-get', 'yinst', 'yum', 'pip3'])
            self.assertListEqual(result.configuration['yinst']['deps'], [])
            self.assertListEqual(result.configuration['yum']['deps'], [])
            self.assertListEqual(result.configuration['pip3']['deps'], [])

    def test__configuration__test__deps(self):
        with InTemporaryDirectory():
            with open('pyproject.toml', 'w') as file_handle:
                file_handle.write(TEST_CONFIG)
            result = Configuration()
            self.assertListEqual(result.configuration['apk']['deps'], ['python3', 'mysql-client'])
            self.assertListEqual(result.configuration['apt-get']['deps'], ['python3', 'mysql-client'])
            self.assertListEqual(result.configuration['yinst']['deps'], ['python36', 'dist_utils'])
            self.assertListEqual(result.configuration['yum']['deps'], ['yahoo_python36;distro_version<"7.5', 'yahoo_python37;distro_version>="7.5"', 'mysql;distro_version<"7"', 'mariadb;distro_version>="7"'])
            self.assertListEqual(result.configuration['pip3']['deps'], [])

    def test__configuration__test__deps__scrwdrivercd_installdeps(self):
        with InTemporaryDirectory():
            with open('pyproject.toml', 'w') as file_handle:
                file_handle.write(TEST_CONFIG.replace('sdv4_installdeps', 'screwdrivercd_installdeps'))
            result = Configuration()
            self.assertListEqual(result.configuration['apk']['deps'], ['python3', 'mysql-client'])
            self.assertListEqual(result.configuration['apt-get']['deps'], ['python3', 'mysql-client'])
            self.assertListEqual(result.configuration['yinst']['deps'], ['python36', 'dist_utils'])
            self.assertListEqual(result.configuration['yum']['deps'], ['yahoo_python36;distro_version<"7.5', 'yahoo_python37;distro_version>="7.5"', 'mysql;distro_version<"7"', 'mariadb;distro_version>="7"'])
            self.assertListEqual(result.configuration['pip3']['deps'], [])

if __name__ == '__main__':
    unittest.main()



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Screwdriver Utility Output helpers
"""
import logging
import os
import shutil
import sys
import textwrap

from termcolor import colored

LOG = logging.getLogger(__name__)



def header(text='', width=None, separator=None, outfile=None, collapse=False):
    """
    Display a textual header message.

    Parameters
    ----------
    text : str
        The text to print/display
    width : int, optional
        The width (text wrap) of the header message.
        This will be the current terminal width as determined
        by the 'stty size' shell command if not specified.
    separator : str, optional
        The character or string to use as a horizontal
        separator.  Will use '=' if one is not specified.
    outfile : File, optional
        The File object to print the header text to.
        This will use sys.stdout if not specified.
    :return:
    """
    if not outfile:
        outfile = sys.stdout

    width = width or shutil.get_terminal_size().columns

    separator = separator or '='

    # Screwdriver buffers stdout and stderr separately.  This can
    # cause output from previous operations to show after our
    # header text.  So we flush the output streams to ensure
    # all existing output is sent/displayed before printing our
    # header.
    sys.stderr.flush()
    outfile.flush()

    if collapse:
        for line in textwrap.wrap(text, width=width-4):
            header_line = separator + ' ' + line + ' ' + separator * (width - 4 - len(line))
            print(header_line, file=outfile)
    else:
        # Print the header text
        horiz = separator * width
        print(horiz, file=outfile)
        print(
            os.linesep.join(textwrap.wrap(text, width=width)),
            file=outfile
        )
        print(horiz, file=outfile)

    # Once again flush our header to the output stream so things
    # show up in the correct order.
    sys.stderr.flush()
    outfile.flush()


def status_message(message: str='', indent: int=0, color: str=''):
    """
    Format a message as an error message

    Parameters
    ----------
    message: str
        The error message

    indent: int
        How far to indent the message

    color: str, optional
        The color to make the message

    Returns
    -------
    str:
        The formatted message
    """
    outstring = textwrap.indent(message, prefix=' ' * indent)
    if color:
        outstring = colored(outstring, color=color)
    return outstring


def print_error(message: str='', indent: int=0, file=sys.stdout):
    """
    Print an error message in color

    Parameters
    ----------
    message: str
        The message to display
    indent: int, optional
        Number of columns to indent, default=0
    file: file
        File type object to send the output to
    """
    print(status_message(message, indent=indent, color='red'), file=file, flush=True)



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for termsimport copy
import base64
import os
import stat
import unittest

from screwdrivercd.screwdriver.github_deploykey import add_github_to_known_hosts, git_key_secret, validate_known_good_hosts
from . import ScrewdriverTestCase


class ScrewdriverGithubDeploykeyTestCase(ScrewdriverTestCase):
    environ_keys = {'GIT_DEPLOY_KEY', 'SD_BUILD', 'SD_BUILD_ID', 'SD_PULL_REQUEST'}

    def setUp(self):
        super().setUp()
        self.known_hosts_filename = os.path.join(self.tempdir.name, '.ssh/known-hosts')

    @unittest.skipIf(not os.path.exists( '/usr/bin/ssh-keyscan'), 'The ssh-keyscan command is missing')
    def test__add_github_to_known_hosts(self):
        result = add_github_to_known_hosts(self.known_hosts_filename)
        self.assertTrue(os.path.exists(self.known_hosts_filename))
        mode = stat.filemode(os.stat(self.known_hosts_filename).st_mode)
        self.assertEqual(mode, '-rw-------')

    @unittest.skip
    def test__validate_known_good_hosts(self):
        add_github_to_known_hosts(self.known_hosts_filename)
        result = validate_known_good_hosts(self.known_hosts_filename)
        self.assertTrue(result)

    def test__git_key_secret(self):
        test_key = b'-----BEGIN RSA PRIVATE KEY-----\nno real key here\n-----END RSA PRIVATE KEY-----\n'
        os.environ['GIT_DEPLOY_KEY'] = base64.b64encode(test_key).decode(errors='ignore')
        result = git_key_secret()
        self.assertEqual(result, test_key)

    def test__git_key_secret__nobegin(self):
        test_key = b'no real key here\n-----END RSA PRIVATE KEY-----\n'
        os.environ['GIT_DEPLOY_KEY'] = base64.b64encode(test_key).decode(errors='ignore')
        result = git_key_secret()
        self.assertEqual(result, test_key)

    def test__git_key_secret__noend(self):
        test_key = b'-----BEGIN RSA PRIVATE KEY-----\nno real key here\n'
        os.environ['GIT_DEPLOY_KEY'] = base64.b64encode(test_key).decode(errors='ignore')
        result = git_key_secret()
        self.assertEqual(result, test_key)

    def test__git_key_secret__empty(self):
        result = git_key_secret()
        self.assertEqual(result, b'')



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
import os
import sys
import unittest
from tempfile import TemporaryDirectory
from screwdrivercd.version.arguments import get_config_default, parse_arguments
from screwdrivercd.version.exceptions import VersionError


class TestArguments(unittest.TestCase):
    orig_argv = None
    origcwd = os.getcwd()

    def setUp(self):
        self.orig_argv = sys.argv
        self.tempdir = TemporaryDirectory()
        os.chdir(self.tempdir.name)

    def tearDown(self):
        if self.orig_argv:
            sys.argv = self.orig_argv
        os.chdir(self.origcwd)
        self.tempdir.cleanup()

    def test__parse_arguments__default(self):
        sys.argv = ['screwdrivercd_version']
        result = parse_arguments()
        self.assertFalse(result.force_update)
        self.assertEqual(result.version_type, 'default')

    def test__parse_arguments__force_update(self):
        sys.argv = ['screwdrivercd_version', '--force_update']
        result = parse_arguments()
        self.assertTrue(result.force_update)
        self.assertEqual(result.version_type, 'default')

    def test__get_config_default__no_setup_cfg(self):
        result = get_config_default('version_type')
        self.assertIsNone(result)

    def test__get_config_default__setup_cfg__key_missing__default(self):
        result = get_config_default('version_type', 'default')
        self.assertEqual(result, 'default')

    def test__parse_arguments__default__setup_cfg__version_type__none(self):
        sys.argv = ['screwdrivercd_version']
        result = parse_arguments()
        self.assertEqual(result.version_type, 'default')

    def test__parse_arguments__default__setup_cfg__version_type(self):
        with open('setup.cfg', 'w') as fh:
            fh.write('[screwdrivercd.version]\nversion_type = utc_date\n')
        sys.argv = ['screwdrivercd_version']
        result = parse_arguments()
        self.assertEqual(result.version_type, 'utc_date')

    def test__parse_arguments__default__setup_cfg__version_type__invalid(self):
        with open('setup.cfg', 'w') as fh:
            fh.write('[screwdrivercd.version]\nversion_type = invalid\n')
        sys.argv = ['screwdrivercd_version']
        with self.assertRaises(VersionError):
            parse_arguments()

    def test__parse_arguments__setup_cfg__update_meta__true(self):
        with open('setup.cfg', 'w') as fh:
            fh.write('[screwdrivercd.version]\nupdate_screwdriver_meta = True\n')
        sys.argv = ['screwdrivercd_version']
        result = parse_arguments()
        self.assertTrue(result.update_meta)

    def test__parse_arguments_legacy__default__setup_cfg__version_type(self):
        with open('setup.cfg', 'w') as fh:
            fh.write('[sdv4.version]\nversion_type = utc_date\n')
        sys.argv = ['screwdrivercd_version']
        result = parse_arguments()
        self.assertEqual(result.version_type, 'utc_date')

    def test__parse_arguments_legacy__default__setup_cfg__version_type__invalid(self):
        with open('setup.cfg', 'w') as fh:
            fh.write('[sdv4.version]\nversion_type = invalid\n')
        sys.argv = ['screwdrivercd_version']
        with self.assertRaises(VersionError):
            parse_arguments()

    def test__parse_arguments_legacy__setup_cfg__update_meta__true(self):
        with open('setup.cfg', 'w') as fh:
            fh.write('[sdv4.version]\nupdate_screwdriver_meta = True\n')
        sys.argv = ['screwdrivercd_version']
        result = parse_arguments()
        self.assertTrue(result.update_meta)




# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
import base64
import os

from screwdrivercd.changelog.generate import changelog_contents, create_first_commit_tag_if_missing, git_tag_dates, write_changelog
from screwdrivercd.repo.release import main as release_main
from screwdrivercd.repo.release import create_release_tag, push_release_tag
from screwdrivercd.utility.environment import env_int, env_bool, interpreter_bin_command, standard_directories


from . import ScrewdriverTestCase


working_config = {
    'setup.py': b"""
from setuptools import setup
setup()
""",
    'setup.cfg': b"""
[metadata]
name=mypyvalidator
version=0.0.0

[options]
packages =
    mypyvalidator

package_dir =
    =src
""",
    'src/mypyvalidator/__init__.py': b"""a: int = 1\n"""
}


class TestPlatformRepoRelease(ScrewdriverTestCase):

    def create_example_repo(self):
        print(os.getcwd())

        os.system('git init')
        os.system('git config user.email "foo@bar.com"')
        os.system('git config user.name "foo the bar"')

        self.write_config_files(working_config)
        self.write_config_files({'changelog.d/1.feature.md': b'Initial commit\n'})
        os.system('git add changelog.d/1.feature.md')
        os.system('git commit -a -m "inital commit"')

        self.write_config_files({'changelog.d/2.feature.md': b'Added another new feature\n'})
        os.system('git add changelog.d/2.feature.md')
        os.system('git commit -a -m "second commit"')
        os.system('git tag -a -m "new tag" v0.0.1')

        self.write_config_files({'changelog.d/3.feature.md': b'Added a second new feature\n'})
        os.system('git add changelog.d/3.feature.md')
        os.system('git commit -a -m "third commit"')
        os.system('git tag -a -m "new tag" v0.1.0')

        self.write_config_files({'changelog.d/4.bugfix.md': b'Fixed the second new feature\n'})
        os.system('git add changelog.d/4.bugfix.md')
        os.system('git commit -a -m "forth commit"')
        os.system('git tag -a -m "new tag" v0.1.1')

    def moc_command(self, command='git', rc=0, stdout='stdout', stderr='stderr', delay=0):
        command_filename = os.path.join(self.tempdir.name, command)
        with open(command_filename, 'w') as fh:
            os.fchmod(fh.fileno(), 0o0755)
            fh.write(f'#!/bin/bash\n')
            if stdout:
                fh.write(f'\ncat << EOF\n{stdout}\nEOF\n')
            if stderr:
                fh.write(f'\ncat << EOF >&2\n{stderr}\nEOF\n')
            if delay:
                fh.write(f'\nsleep {delay}\n')
            fh.write(f'exit {rc}\n')
        return command_filename

    def test__release__main__publish_disabled(self):
        os.environ['PUBLISH'] = 'False'
        result = release_main()
        self.assertEqual(result, 0)

    def test__release__main__package_tag__disabled(self):
        os.environ['PUBLISH'] = 'True'
        os.environ['PACKAGE_TAG'] = 'False'
        result = release_main()
        self.assertEqual(result, 0)

    def test__release__main__git_deploy_key__missing(self):
        os.environ['PUBLISH'] = 'True'
        os.environ['PACKAGE_TAG'] = 'True'
        result = release_main()
        self.assertEqual(result, 0)

    def test__release__create_tag(self):
        self.create_example_repo()
        create_release_tag(version='999.999.999')
        result = git_tag_dates()
        self.assertIn('v999.999.999', result.keys())

    def test__release__create_tag__message(self):
        self.create_example_repo()
        create_release_tag(version='999.999.999', message='test message')
        result = git_tag_dates()
        self.assertIn('v999.999.999', result.keys())

    def test__release__create_tag__git_error(self):
        self.create_example_repo()
        create_release_tag(version='999.999.999', git_command=self.moc_command(command='git', rc=999))

    def test__release_push_tag(self):
        self.create_example_repo()
        push_release_tag(git_command=self.moc_command(command='git', rc=0))

    def test__release_push_tag__cmd_missing(self):
        self.create_example_repo()
        push_release_tag(git_command=self.moc_command(command='gitbad', rc=1))

    def test__release_push_tag__fail(self):
        self.create_example_repo()
        push_release_tag(git_command=self.moc_command(command='git', rc=1))

    def test__release_push_tag__timeout(self):
        self.create_example_repo()
        push_release_tag(git_command=self.moc_command(command='git', rc=0, delay=61), timeout=1)

    def test__main(self):
        test_key = b'-----BEGIN RSA PRIVATE KEY-----\nno real key here\n-----END RSA PRIVATE KEY-----\n'
        os.environ['GIT_DEPLOY_KEY'] = base64.b64encode(test_key).decode(errors='ignore')
        self.create_example_repo()
        release_main(meta_command=self.moc_command(command='meta', stdout='999.999.999'))
        result = git_tag_dates()
        self.assertIn('v999.999.999', result.keys())

    def test__main__no_version(self):
        test_key = b'-----BEGIN RSA PRIVATE KEY-----\nno real key here\n-----END RSA PRIVATE KEY-----\n'
        os.environ['GIT_DEPLOY_KEY'] = base64.b64encode(test_key).decode(errors='ignore')
        self.create_example_repo()
        release_main(meta_command=self.moc_command(command='meta', stdout='null'))



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""Functions for working with the setup.py file"""
import configparser
import logging


LOGGER_NAME = 'setup' if __name__ == '__main__' else __name__
LOG = logging.getLogger(LOGGER_NAME)


def setupcfg_has_metadata(setup_cfg_filename='setup.cfg'):
    """Parse the setup.cfg and return True if it has a metadata section"""
    config = configparser.ConfigParser()
    config.read(setup_cfg_filename)
    if 'metadata' in config.sections():
        return True
    return False



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for termsimport copy
import base64
import os
import stat
import unittest

from screwdrivercd.changelog.generate import changelog_contents, create_first_commit_tag_if_missing, git_tag_dates, write_changelog
from screwdrivercd.changelog.generate import main as changelog_generate_main
from . import ScrewdriverTestCase


working_config = {
    'setup.py': b"""
from setuptools import setup
setup()
""",
    'setup.cfg': b"""
[metadata]
name=mypyvalidator
version=0.0.0

[options]
packages =
    mypyvalidator

package_dir =
    =src
""",
    'src/mypyvalidator/__init__.py': b"""a: int = 1\n"""
}


class ScrewdriverChangelogTestCase(ScrewdriverTestCase):

    def create_example_repo(self):
        print(os.getcwd())

        os.system('git init')
        os.system('git config user.email "foo@bar.com"')
        os.system('git config user.name "foo the bar"')

        self.write_config_files(working_config)
        self.write_config_files({'changelog.d/1.feature.md': b'Initial commit\n'})
        os.system('git add changelog.d/1.feature.md')
        os.system('git commit -a -m "inital commit"')

        self.write_config_files({'changelog.d/2.feature.md': b'Added another new feature\n'})
        os.system('git add changelog.d/2.feature.md')
        os.system('git commit -a -m "second commit"')
        os.system('git tag -a -m "new tag" v0.0.1')

        self.write_config_files({'changelog.d/3.feature.md': b'Added a second new feature\n'})
        os.system('git add changelog.d/3.feature.md')
        os.system('git commit -a -m "third commit"')
        os.system('git tag -a -m "new tag" v0.1.0')

        self.write_config_files({'changelog.d/4.bugfix.md': b'Fixed the second new feature\n'})
        os.system('git add changelog.d/4.bugfix.md')
        os.system('git commit -a -m "forth commit"')
        os.system('git tag -a -m "new tag" v0.1.1')

    def test__changelog_generate_main(self):
        self.create_example_repo()

        changelog_generate_main()

        self.assertTrue(os.path.exists('artifacts/reports/changelog/changelog.md'))
        with open('artifacts/reports/changelog/changelog.md') as fh:
            changelog_contents = fh.read()

        print(changelog_contents)

        self.assertNotIn('# mypyvalidator first_commit (', changelog_contents)
        self.assertNotIn('# mypyvalidator new (', changelog_contents)
        self.assertIn('# mypyvalidator v0.0.1 (', changelog_contents)
        self.assertIn('# mypyvalidator v0.1.0 (', changelog_contents)
        self.assertIn('# mypyvalidator v0.1.1 (', changelog_contents)

    def test__changelog_generate_header(self):
        self.create_example_repo()
        self.write_config_files({'changelog.d/HEADER.md': b'# Changelog header\n'})

        changelog_generate_main()

        self.assertTrue(os.path.exists('artifacts/reports/changelog/changelog.md'))
        with open('artifacts/reports/changelog/changelog.md') as fh:
            changelog_contents = fh.read()

        print(changelog_contents)

        self.assertNotIn('# mypyvalidator first_commit (', changelog_contents)
        self.assertNotIn('# mypyvalidator new (', changelog_contents)
        self.assertIn('# mypyvalidator v0.0.1 (', changelog_contents)
        self.assertIn('# mypyvalidator v0.1.0 (', changelog_contents)
        self.assertIn('# mypyvalidator v0.1.1 (', changelog_contents)
        self.assertIn('# Changelog header', changelog_contents)

    def test__changelog_generate_footer(self):
        self.create_example_repo()
        self.write_config_files({'changelog.d/FOOTER.md': b'# Changelog footer\n'})

        changelog_generate_main()

        self.assertTrue(os.path.exists('artifacts/reports/changelog/changelog.md'))
        with open('artifacts/reports/changelog/changelog.md') as fh:
            changelog_contents = fh.read()

        print(changelog_contents)

        self.assertNotIn('# mypyvalidator first_commit (', changelog_contents)
        self.assertNotIn('# mypyvalidator new (', changelog_contents)
        self.assertIn('# mypyvalidator v0.0.1 (', changelog_contents)
        self.assertIn('# mypyvalidator v0.1.0 (', changelog_contents)
        self.assertIn('# mypyvalidator v0.1.1 (', changelog_contents)
        self.assertIn('# Changelog footer', changelog_contents)

    def test__changelog_generate_main__releases_all(self):
        os.environ['CHANGELOG_RELEASES'] = 'all'
        self.create_example_repo()

        changelog_generate_main()

        self.assertTrue(os.path.exists('artifacts/reports/changelog/changelog.md'))
        with open('artifacts/reports/changelog/changelog.md') as fh:
            changelog_contents = fh.read()

        print(changelog_contents)

        self.assertNotIn('# mypyvalidator first_commit (', changelog_contents)
        self.assertNotIn('# mypyvalidator new (', changelog_contents)
        self.assertIn('# mypyvalidator v0.0.1 (', changelog_contents)
        self.assertIn('# mypyvalidator v0.1.0 (', changelog_contents)
        self.assertIn('# mypyvalidator v0.1.1 (', changelog_contents)

    def test__changelog_generate_main__nofilter_nonversons(self):
        os.environ['CHANGELOG_ONLY_VERSION_TAGS'] = 'False'

        self.create_example_repo()

        self.write_config_files({'changelog.d/6.bugfix.md': b'Make sure version filter can be turned off\n'})
        os.system('git add changelog.d/6.bugfix.md')
        os.system('git commit -a -m "sixth commit"')
        os.system('git tag -a -m "latest tag" latest')

        changelog_generate_main()

        self.assertTrue(os.path.exists('artifacts/reports/changelog/changelog.md'))
        with open('artifacts/reports/changelog/changelog.md') as fh:
            changelog_contents = fh.read()

        os.system('git log --date-order --tags --simplify-by-decoration --pretty=format:"%ct|%D"')
        print(changelog_contents)

        self.assertNotIn('# mypyvalidator first_commit (', changelog_contents)
        self.assertIn('# mypyvalidator latest (', changelog_contents)
        self.assertIn('# mypyvalidator v0.0.1 (', changelog_contents)
        self.assertIn('# mypyvalidator v0.1.0 (', changelog_contents)
        self.assertIn('# mypyvalidator v0.1.1 (', changelog_contents)

    def test_generate_main__single_version(self):
        os.environ['CHANGELOG_RELEASES'] = 'v0.1.0'
        self.create_example_repo()

        changelog_generate_main()

        self.assertTrue(os.path.exists('artifacts/reports/changelog/changelog.md'))
        with open('artifacts/reports/changelog/changelog.md') as fh:
            changelog_contents = fh.read()

        print(changelog_contents)

        self.assertNotIn('# mypyvalidator v0.0.1 (', changelog_contents)
        self.assertIn('# mypyvalidator v0.1.0 (', changelog_contents)
        self.assertNotIn('# mypyvalidator v0.1.1 (', changelog_contents)

    def test_generate_main__selected_versions(self):
        os.environ['CHANGELOG_RELEASES'] = 'v0.0.1, v0.1.0'
        self.create_example_repo()

        changelog_generate_main()

        self.assertTrue(os.path.exists('artifacts/reports/changelog/changelog.md'))
        with open('artifacts/reports/changelog/changelog.md') as fh:
            changelog_contents = fh.read()

        print(changelog_contents)

        self.assertIn('# mypyvalidator v0.0.1 (', changelog_contents)
        self.assertIn('# mypyvalidator v0.1.0 (', changelog_contents)
        self.assertNotIn('# mypyvalidator v0.1.1 (', changelog_contents)

    def test_generate_main__only_stable_true(self):
        os.environ['CHANGELOG_ONLY_STABLE_RELEASES'] = 'True'
        self.create_example_repo()
        self.write_config_files({'changelog.d/10.feature.md': b'Pre-release commit\n'})
        os.system('git add changelog.d/10.feature.md')
        os.system('git commit -a -m "inital commit"')
        os.system('git tag -a -m "pre release" v0.1.10a1')

        changelog_generate_main()

        self.assertTrue(os.path.exists('artifacts/reports/changelog/changelog.md'))
        with open('artifacts/reports/changelog/changelog.md') as fh:
            changelog_contents = fh.read()

        print(changelog_contents)

        self.assertIn('# mypyvalidator v0.0.1 (', changelog_contents)
        self.assertIn('# mypyvalidator v0.1.0 (', changelog_contents)
        self.assertNotIn('# mypyvalidator v0.1.10a1 (', changelog_contents)

    def test_changelog_contents_with_header_and_footer(self):
        header_content = b"# This is the header"
        footer_content = b"# This is the footer"
        self.create_example_repo()
        self.write_config_files({'changelog.d/HEADER.md': header_content, 'changelog.d/FOOTER.md': footer_content})
        os.system('git add changelog.d/HEADER.md')
        os.system('git add changelog.d/FOOTER.md')
        os.system('git commit -a -m "inital commit"')
        os.system('git tag -a -m "pre release" v0.2.10')

        changelog_generate_main()

        self.assertTrue(os.path.exists('artifacts/reports/changelog/changelog.md'))
        with open('artifacts/reports/changelog/changelog.md') as fh:
            changelog_contents = fh.read()

        print(changelog_contents)

        self.assertIn("This is the header", changelog_contents)
        self.assertIn("This is the footer", changelog_contents)

    def test_generate_main__only_stable_false(self):
        os.environ['CHANGELOG_ONLY_STABLE_RELEASES'] = 'False'
        self.create_example_repo()
        self.write_config_files({'changelog.d/10.feature.md': b'Pre-release commit\n'})
        os.system('git add changelog.d/10.feature.md')
        os.system('git commit -a -m "inital commit"')
        os.system('git tag -a -m "pre release" v0.1.10a1')

        changelog_generate_main()

        self.assertTrue(os.path.exists('artifacts/reports/changelog/changelog.md'))
        with open('artifacts/reports/changelog/changelog.md') as fh:
            changelog_contents = fh.read()

        print(changelog_contents)

        self.assertIn('# mypyvalidator v0.0.1 (', changelog_contents)
        self.assertIn('# mypyvalidator v0.1.0 (', changelog_contents)
        self.assertIn('# mypyvalidator v0.1.10a1 (', changelog_contents)

    def test_create_first_commit_tag_if_missing__missing(self):
        self.create_example_repo()

        before = git_tag_dates()
        self.assertNotIn('first_commit', before.keys())

        create_first_commit_tag_if_missing()

        after = git_tag_dates()
        self.assertIn('first_commit', after.keys())

    def test_create_first_commit_tag_if_missing__present(self):
        self.create_example_repo()

        create_first_commit_tag_if_missing()

        before = git_tag_dates()
        self.assertIn('first_commit', before.keys())

        create_first_commit_tag_if_missing()

        after = git_tag_dates()
        self.assertIn('first_commit', after.keys())

    def test_generate_main__invalid_type(self):
        self.create_example_repo()
        self.write_config_files({'changelog.d/5.moose.md': b'Bad changelog entry\n'})
        os.system('git add changelog.d/5.moose.md')
        os.system('git commit -a -m "fifth commit"')
        os.system('git tag v0.2.0')

        changelog_generate_main()

        self.assertTrue(os.path.exists('artifacts/reports/changelog/changelog.md'))
        with open('artifacts/reports/changelog/changelog.md') as fh:
            changelog_contents = fh.read()

        print(changelog_contents)

        self.assertNotIn('# mypyvalidator v0.2.0 (', changelog_contents)



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""Functions for command script"""
# The logging_basicConfig has to be run before other imports because some modules we use log output on import
# pylint: disable=wrong-import-order, wrong-import-position
from ..screwdriver.environment import logging_basicConfig,  update_job_status
logging_basicConfig(check_prefix='VERSION')

import logging
import sys
from .arguments import parse_arguments
from .setup import setupcfg_has_metadata
from .version_types import versioners


LOG_NAME = 'platform_version' if __name__ == '__main__' else __name__
LOG = logging.getLogger(LOG_NAME)


def main():
    """Run the tool from the command line"""
    args = parse_arguments()

    if args.force_update or setupcfg_has_metadata():
        versioner = versioners[args.version_type]
        print(f'Updating version using the {versioner.name} version plugin', flush=True)
        version = versioner(ignore_meta_version=args.ignore_meta, update_sdv4_meta=args.update_meta, link_to_project=args.link_to_project)

        version.update_setup_cfg_metadata()
        if args.update_meta:
            version.update_meta_version()
            update_job_status(status='SUCCESS', message=f'version=={version}')
            print(f'Version in setup.cfg and screwdriver v4 metadata package.version updated to: {version}', flush=True)
        else:
            print(f'Version in setup.cfg updated to: {version}')
    else:
        print("The 'setup.cfg' file does not have a metadata section with a version in it, not modifying it", file=sys.stderr, flush=True)


if __name__ == '__main__':
    main()



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Screwdriver integration utilities
"""
__all__ = ['environment', 'github_deploykey', 'metadata']



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Package installer classes
"""
from .apk import ApkInstaller
from .apt import AptInstaller
from .brew import BrewInstaller
from .pip3 import PipInstaller
from .yum import YumInstaller


__all__ = ['apk', 'apt', 'brew', 'pip3', 'yum']
install_plugins = {
    'apk': ApkInstaller,
    'apt-get': AptInstaller,
    'brew': BrewInstaller,
    'pip3': PipInstaller,
    'yum': YumInstaller
}



"""
MkDocs documentation generation plugin
"""
import os
import subprocess  # nosec
import tempfile

from typing import List

from ..plugin import DocumentationPlugin
from ...utility.environment import interpreter_bin_command


class MkDocsDocumentationPlugin(DocumentationPlugin):
    """
    screwdrivercd.documentation plugin for mkdocs documentation
    """
    name = 'mkdocs'
    build_output_dir = ''
    # Should cover configuration files supported by https://github.com/mkdocs/mkdocs/blob/3db7b8c9c2473e763fc77525ee17c42e965d3b91/mkdocs/config/base.py#L306
    possible_config_files = ['mkdocs.yml', 'docs/mkdocs.yml', 'mkdocs.yaml', 'docs/mkdocs.yaml']

    def __init__(self, *args, **kwargs):  # pragma: no cover
        super().__init__(*args, **kwargs)
        self.build_command = [interpreter_bin_command(), '-m', 'mkdocs', 'build', '--config-file', self.config_file, '--site-dir', self.build_dir]

    @property
    def config_file(self) -> str:
        for cfile in self.possible_config_files:
            if os.path.exists(cfile):
                return cfile
        return self.possible_config_files[0]

    @property
    def documentation_is_present(self) -> bool:
        """
        Returns
        =======
        bool:
            Check that the documentation source is present, Returns True if it is, False otherwise
        """
        if os.path.exists(self.config_file):
            return True
        return False


class MkDocsDocumentationVenvPlugin(MkDocsDocumentationPlugin):
    """
    screwdrivercd.documentation plugin for mkdocs documentation in a python venv
    """
    name = 'mkdocs_venv'
    venv_dir: str = ''
    default_requirements: List[str] = ['mkdocs', 'markdown', 'pymdown-extensions', 'markdown-include', 'mkdocs-material', 'pygments']

    def build_setup(self):
        """
        Set up a temporary Python virtualenv before running the build
        """
        self.tempdir = tempfile.TemporaryDirectory()
        self.venv_dir = os.path.join(self.tempdir.name, 'mkdocsvenv')
        self.venv_bin = os.path.join(self.venv_dir, 'bin')
        self.venv_python = os.path.join(self.venv_bin, 'python3')

        subprocess.check_call([interpreter_bin_command(), '-m', 'venv', self.venv_dir])  # nosec

        if os.path.exists('documentation_mkdocs_requirements.txt'):
            requirements_file = 'documentation_mkdocs_requirements.txt'
        else:
            requirements_file = os.path.join(self.tempdir.name, 'documentation_mkdocs_requirements.txt')
            with open(requirements_file, 'w') as fh:
                fh.write('\n'.join(self.default_requirements))

        if os.path.exists(requirements_file):
            self._run_command([self.venv_python, '-m', 'pip', 'install', '-r', requirements_file], log_filename=self.build_log_filename)  # nosec

        self.build_command = [self.venv_python, '-m', 'mkdocs', 'build', '--config-file', self.config_file, '--site-dir', self.build_dir]

    def build_cleanup(self):
        """
        Clean up the temporary virtualenv
        """
        self.tempdir.cleanup()



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""Base Installer Functionality"""
import copy
import logging
import os
import shutil
import subprocess  # nosec - All subprocess calls use full path
import sys
from typing import Dict, Optional, List

from termcolor import colored
from .config import Configuration
from .requirement import Requirement


LOG = logging.getLogger(__name__)


class Installer():
    """Generic Package Installer
    """
    name: str = 'echo'
    """:str: The name of the installer class"""

    install_command: List[str] = ['echo', 'fake', 'install']
    """:obj:`list` of :obj:`str`: The base install command list"""

    bin_dir: Optional[str] = None
    """str: The directory that contains the install command binary, or None to search the PATH"""

    install_command_path: Optional[List[str]] = []
    """:obj:`list` of :obj:`str`: A list of directories to search for the install command"""

    use_system_path: bool = False
    """bool: If True, search the system path if the command cannot be found in the install_command_path"""

    supports_repositories: bool = False
    """bool: If True the installer supports adding repositories"""

    default_repos: Optional[Dict[str, str]] = {}
    """A dictionary of repo name, repo url to enable by default"""

    config_section: str = 'echo'
    """str: The name of the config section that this installer configuration is found in"""

    print_output: bool = False
    """bool: If True, print the output to stdout instead of logging"""

    print_error_output: bool = True
    """bool: If True, print the output to stdout instead of logging when an error occurs"""

    exit_on_missing: bool = False
    """bool: If True, terminate installation immediately if a dependency is missing"""

    def __init__(self, dry_run: bool = False, bin_dir: Optional[str] = None):
        """
        Generic Package Installer

        Parameters
        ----------
        bin_dir: str, optional
            The directory path that contains the binaries

        dry_run: bool, optional
            If True, Don't execute packaging commands, default=False
        """
        self.dry_run = dry_run

        if bin_dir:
            self.bin_dir = bin_dir


        self.config = Configuration()
        self.find_install_command()
        self._handle_custom_settings()

    @property
    def deps_config_keys(self):
        """
        All of the configuration keys that have dependencies for this installer

        Returns
        -------
        list of str
            The settings that contain dependencies to be installed
        """
        # This is a property to allow classes derived from this one to
        # add dependencies to the list returned dynamically depending
        # on system state.
        return ['deps']

    @property
    def has_dependencies(self) -> bool:  # pragma: no cover
        """
        Check if the installer has dependencies to install in the current environment.

        Returns
        =======
        bool:
            True if there are dependencies to install, False otherwise
        """
        for config_key in self.deps_config_keys:
            dependencies = self.config.configuration[self.config_section][config_key]
            if dependencies:
                dependencies = self.filter_environment_markers(dependencies)
                if dependencies:
                    return True
        return False

    @property
    def is_supported(self) -> bool:  # pragma: no cover
        """
        Check for the installer utility and return True if it's present, False otherwise

        Returns
        =======
        bool:
            True if the installer is supported on the host system, False otherwise
        """
        self.find_install_command()
        if os.path.exists(self.install_command[0]):
            return True
        return False

    @property
    def plugin_configuration(self):  # pragma: no cover
        """
        Return the plugin configuration dictionary if it is found

        Returns
        -------
        dict:
            Plugin configuration dictionary
        """
        if not self.config:
            return {}
        if not self.config.configuration:
            return {}
        return self.config.configuration.get(self.config_section, {})

    def _handle_custom_settings(self):
        """
        Allows classes derived from this one to handle custom config options outside of the dependency sections
        Parameters
        ----------
        config
            parsed config dictionary

        Returns
        -------
        """

    def add_repos(self):  # pragma: no cover
        """
        Add repositories to the host configuration
        """
        if not self.supports_repositories:
            LOG.debug(f'Install plugin {self.config_section} does not support repositories')
            return

        repos = copy.copy(self.default_repos)
        repos.update(self.plugin_configuration.get('repos', {}))

        LOG.debug('Adding package repositories')
        for repo_name, repo_url in repos.items():
            repo_url_env = repo_url.split(';')
            if len(repo_url_env) > 1:
                repo_url_req = Requirement(f'{repo_name};{repo_url_env[-1]}')
                LOG.debug(f'Evaluating the repository environment marker {repo_name};{repo_url_env[-1]} == {repo_url_req.env_matches}')
                if not repo_url_req.env_matches:
                    LOG.debug(f'Filtered {repo_name!r} as it is not supported in this environment')
                    continue
                repo_url = repo_url_env[0]
            if self.print_output:
                print(f'Enabling {repo_name!r} repo: ', end='', flush=True)
            else:
                LOG.debug(f'Enabling {repo_name!r} repo')
            self.add_repo(repo_name=repo_name, repo_url=repo_url)

    def add_repo(self, repo_name, repo_url):  # pragma: no cover
        """
        Add a repository to the host configuration

        Parameters
        ----------
        repo_name: str
            The human readable name string for the repository

        repo_url: str
            The url of the repository to add
        """

    def determine_bin_directory(self):
        """
        Determine the bin_dir for the default install command based on the environment
        """
        if self.bin_dir or self.install_command[0].startswith('/'):
            LOG.debug(f'The command bin_dir {self.bin_dir} is already set')
            return

        if self.plugin_configuration and self.plugin_configuration.get('bin_dir', None):  # pragma: no cover
            LOG.debug('Setting the configuartion directory from the configuration')
            self.bin_dir = self.plugin_configuration['bin_dir']

        # LOG.debug(f'Searching for the install command in {self.install_command_path}')
        for directory in self.install_command_path:
            filename = os.path.join(directory, self.install_command[0])
            if os.path.exists(filename):
                # LOG.debug(f'Found install command in install_command_path {directory!r} directory')
                self.bin_dir = directory
                return

        if self.use_system_path:
            install_binary = shutil.which(self.install_command[0])
            if install_binary:
                self.bin_dir = os.path.dirname(install_binary)
                LOG.debug(f'Found install command in system path {self.bin_dir}')
                return

    def find_install_command(self):
        """
        Find the install command binary to use and update the install command
        """
        self.determine_bin_directory()
        if self.install_command[0].startswith('/'):
            return

        if self.bin_dir:
            full_command = os.path.abspath(os.path.join(self.bin_dir, self.install_command[0]))
            self.install_command[0] = full_command
        else:
            if self.use_system_path:
                LOG.debug(f'Finding the install command in the system PATH')
                install_command = shutil.which(self.install_command[0])
                if install_command:
                    self.install_command[0] = install_command
                return

    def filter_environment_markers(self, dependencies: List[str]) -> List[str]:
        """
        Remove any dependencies that don't match environment markers for the current environment
        """
        new_dependencies = []
        for dependency in dependencies:
            req = Requirement(dependency)
            if not req.env_evals:
                new_dependencies.append(dependency)
                continue

            if req.env_matches:
                dependency = dependency.split(';')[0]
                new_dependencies.append(dependency)
                continue

            LOG.debug(f'Filtered dependency {dependency} due to the environment marker')
        return new_dependencies

    def install_dependencies(self) -> List[str]:
        """
        Install all of the dependencies using the install command
        """
        if not self.has_dependencies:
            LOG.debug(f'No {self.name!r} dependencies to install')
            return []

        self.add_repos()
        self.update_index()

        installed: List[str] = []
        invalid: List[str] = []
        for config_key in self.deps_config_keys:
            LOG.debug(f'Looking for dependencies in [tool.sdv4_installdeps.{self.config_section}] {config_key}')
            dependencies = self.config.configuration[self.config_section][config_key]
            if dependencies:
                dependencies = self.filter_environment_markers(dependencies)
                # LOG.debug(f'Processing dependencies {dependencies!r}')
                invalid += self.invalid_dependencies(dependencies, config_key=config_key)
                if invalid:
                    if self.print_error_output:  # pragma: no cover
                        print(colored('Invalid %r dependencies %r specified' % (config_key, invalid), 'red'), flush=True)
                    else:
                        LOG.error(colored('Invalid %r dependencies %r specified' % (config_key, invalid), 'red'))
                    if self.exit_on_missing:
                        break
                    dependencies = list(set(dependencies) - set(invalid))

                if dependencies:
                    if self.print_output:
                        print(f'Installing dependencies {dependencies!r}')
                    else:
                        LOG.debug(f'Installing dependencies: {dependencies!r}')
                    installed += self.install(dependencies, config_key=config_key)
        return installed

    def install(self, dependencies, config_key=None):
        """
        Install a list of dependencies

        Parameters
        ----------
        dependencies: list
            List of the dependencies to install

        Returns
        -------
        list:
            Dependencies installed
        """
        command = self.install_command + self.install_arguments(config_key=config_key) + dependencies
        if self.dry_run:
            LOG.debug('Dry Run: %r', ' '.join(command))
            return dependencies
        if self.print_output:
            print('Running command: %r' % ' '.join(command))
        else:  # pragma: no cover
            LOG.debug('Running command: %r', ' '.join(command))
        try:
            output = subprocess.check_output(command, stderr=subprocess.STDOUT)  # nosec - All subprocess calls use full path
        except subprocess.CalledProcessError as error:  # pragma: no cover
            if self.print_error_output:
                print(colored(f'Install command {" ".join(command)!r} failed', "red"), flush=True)
                print(error.stdout.decode().strip())
                if self.exit_on_missing:
                    sys.exit(1)
            else:
                LOG.error('Install command failed')
                LOG.error(error.stdout.decode.strip())
            return []

        if self.print_output:  # pragma: no cover
            print(output.decode().strip())
        else:  # pragma: no cover
            LOG.debug(output.decode().strip())

        return dependencies

    def validate_dependency(self, dependency):
        """
        Validate a dependency is valid

        Parameters
        ----------
        dependency: str
            Dependency to validate

        Returns
        -------
        bool
            True if valid
        """
        return True

    def invalid_dependencies(self, dependencies, config_key=None):
        """
        Check that all dependencies in the list are valid

        Parameters
        ----------
        dependencies: list of str:
            The dependency list to validate

        Returns
        -------
        list of str:
            Dependencies that are invalid
        """
        invalid = []
        for depend in dependencies:
            if self.validate_dependency(depend) is False:
                invalid.append(depend)
        return invalid

    def install_arguments(self, config_key=None):
        """
        Return extra command line arguments list based on the config_key

        Returns
        -------
        list of str:
            Extra arguments to add
        """
        return []

    def update_index(self):
        """
        Method to update the package index
        """



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
test_setup
----------

Tests for the 'yahoo.platform_version.setup' module.
"""
import os
import unittest
from tempfile import TemporaryDirectory
from screwdrivercd.version.setup import setupcfg_has_metadata


class TestSetup(unittest.TestCase):

    environ_backup = None
    origcwd = os.getcwd()

    def setUp(self):
        # Save os.environ values we may need to change
        environ_backup = {}
        for key in ['BUILD_NUMBER', 'JOB_TYPE', 'SD_PULL_REQUEST']:
            environ_backup[key] = os.environ.get(key, None)
        self.tempdir = TemporaryDirectory()
        os.chdir(self.tempdir.name)

    def tearDown(self):
        # Restore the os.environ values in case they were changed
        if self.environ_backup:
            for key, value in self.environ_backup:
                if value is None:
                    del os.environ[key]
                else:
                    os.environ[key] = value
        self.environ_backup = None
        os.chdir(self.origcwd)
        self.tempdir.cleanup()

    def test__setupcfg_has_metadata__true(self):
        with open('setup.cfg', 'w') as setup_cfg_handle:
            setup_cfg_handle.write('[metadata]\nversion = 1.2.3\n\n')
        self.assertTrue(setupcfg_has_metadata())

    def test__setupcfg_has_metadata__false(self):
        with open('setup.cfg', 'w') as setup_cfg_handle:
            setup_cfg_handle.write('[foo]\nversion = 1.2.3\n\n')
        self.assertFalse(setupcfg_has_metadata())

    def test__setupcfg_has_metadata__missing(self):
        self.assertFalse(setupcfg_has_metadata())


if __name__ == '__main__':
    unittest.main()



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
screwdrivercd.changelog module entrypoint
"""
import sys
from .generate import main


sys.exit(main())



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
screwdrivercd version module
"""
from typing import List
from importlib.metadata import version, PackageNotFoundError


__all__: List[str] = ['arguments', 'cli', 'exceptions', 'setup', 'version_types']
__copyright__: str = "Copyright 2019, Oath Inc."

__version__: str = "0.0.0"
try:
    __version__ = version("screwdrivercd")
except PackageNotFoundError:
    pass



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Code style validation wrapper for screwdrivercd

This wrapper runs the pycodestyle validation tool.
"""
# The logging_basicConfig has to be run before other imports because some modules we use log output on import
# pylint: disable=wrong-import-order, wrong-import-position
from ..screwdriver.environment import logging_basicConfig
logging_basicConfig(check_prefix='STYLE_CHECK')

import logging
import os
import subprocess  # nosec
import sys

from pypirun.cli import interpreter_parent
from termcolor import colored

from ..utility import create_artifact_directory
from ..utility.environment import ins_filename
from ..utility.output import print_error
from ..utility.package import PackageMetadata, package_srcdir


logger_name = 'validate_style' if __name__ == '__main__' else __name__
logger = logging.getLogger(logger_name)


def validate_with_codestyle(report_dir):
    """Run the codestyle command directly to do the validation"""

    package_metadata = PackageMetadata()
    package_name = package_metadata.metadata['name']
    src_dir = package_srcdir()

    parent_interpreter = interpreter_parent(sys.executable)
    interpreter = os.environ.get('BASE_PYTHON', parent_interpreter)
    bin_dir = os.path.dirname(interpreter)

    pycodestyle_command = os.path.join(bin_dir, 'pycodestyle')
    if not os.path.exists(pycodestyle_command):  # pragma: no cover
        bin_dir = os.path.dirname(sys.executable)
        pycodestyle_command = os.path.join(bin_dir, 'pycodestyle')
        if not os.path.exists(pycodestyle_command):
            bin_dir = os.path.dirname(parent_interpreter)
            pycodestyle_command = os.path.join(bin_dir, 'pycodestyle')
            if not os.path.exists(pycodestyle_command):
                pycodestyle_command = 'pycodestyle'

    # Generate the command line from the environment settings
    command = [pycodestyle_command]

    # Add extra arguments
    extra_args = os.environ.get('CODESTYLE_ARGS', '')
    if extra_args:
        command += extra_args.split()

    if src_dir and src_dir not in ['.']:
        target = src_dir
    else:
        # Add targets
        target = ''
        src_dir = '.'
        if hasattr(package_metadata, 'packages'):
            for package in package_metadata.packages:
                package_path = os.path.join(src_dir, package)
                if os.path.exists(package_path):
                    target = package_path
                    break

        if not target:
            if src_dir not in ['', '.'] and src_dir != package_name:
                target = os.path.join(src_dir, package_name.replace('.', '/'))
            else:
                target = package_name.replace('.', '/')

        print(f'target: {target}')
        target = ins_filename(target)
        if not target:
            print_error(f'ERROR: Unable to find package directory for package {package_name!r}, target directory {target!r} does not exist')
            return 1
    command.append(target)

    print('-' * 90 + '\nRunning:', ' '.join(command) + '\n' + '-' * 90, flush=True)
    rc = 0
    try:
        output = subprocess.check_output(command, stderr=subprocess.STDOUT)  # nosec
    except subprocess.CalledProcessError as error:  # pragma: no cover
        rc = error.returncode
        output = error.output

    os.makedirs(report_dir, exist_ok=True)

    text_report = os.path.join(report_dir, 'codestyle.txt')

    with open(text_report, 'wb') as fh:
        for line in output.split(b'\n'):
            textline = line.decode(errors='ignore').strip()
            fh.write(line)
            if 'error:' in textline:  # pragma: no cover
                print(colored(textline, 'red'), flush=True)
            else:
                print(textline, flush=True)

    return rc


def validate_codestyle():
    """
    Run the codestyle validator tool and store the output

    Returns
    -------
    int:
        Exit returncode from the style validation command
    """
    logging_basicConfig()

    # Set the status message
    # update_job_status(status='SUCCESS', message='Checking code style')

    # Make sure the report directory exists
    artifacts_dir = os.environ.get('SD_ARTIFACTS_DIR', '')
    report_dir = os.path.join(artifacts_dir, 'reports/style_validation')
    create_artifact_directory(report_dir)

    rc = validate_with_codestyle(report_dir=report_dir)

    if rc == 0:
        print(colored('OK: Code style validation successful', 'green'), flush=True)

    if rc > 0:
        print(colored('ERROR: Code style check failed', 'red'), file=sys.stderr, flush=True)
        return rc

    return rc


def main() -> int:
    """
    Codestyle check runner utility command line entry point

    Returns
    -------
    int:
        Returncode from running the check
    """
    return validate_codestyle()


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
"""
Screwdrivercd wrappers and utilities to perform code validation
"""
__all__ = ['validate_dependencies', 'validate_package_quality', 'validate_style', 'validate_type', 'validate_unittest']



#!/usr/bin/env python
# Copyright 2019, Oath Inc.
# Licensed under the terms of the Apache 2.0 license.  See the LICENSE file in the project root for terms
import os
import tempfile
import unittest.mock

import distro
from screwdrivercd.installdeps.installers.yum import YumInstaller
from screwdrivercd.utility.contextmanagers import InTemporaryDirectory


CONFIG_FILE = 'pyproject.toml'
TEST_CONFIG = f'''[build-system]
# Minimum requirements for the build system to execute.
requires = ["setuptools", "wheel"]  # PEP 508 specifications.

[tool.sdv4_installdeps]
    install = ['apk', 'apt-get', 'yinst', 'yum', 'pip3']

    [tool.sdv4_installdeps.yum]
        deps = [
            'python36;distro_version=="{distro.version()}"',
            'python37;distro_version>="7.5"',
            'mysql;distro_version<"7"',
            'mariadb;distro_version>="7"'
        ]

'''


class TestYum(unittest.TestCase):
    original_environ = None

    def setUp(self):
        self._cwd = os.getcwd()
        super().setUp()
        self.original_environ = os.environ
        self.tempdir = tempfile.TemporaryDirectory()
        os.chdir(self.tempdir.name)
        with open(CONFIG_FILE, 'w') as config_handle:
            config_handle.write(TEST_CONFIG)
        print(f'Using configuration: \n{TEST_CONFIG}')

    def tearDown(self):
        super().tearDown()
        if self.original_environ:
            os.environ = self.original_environ
            self.original_environ = None
        os.chdir(self._cwd)
        self.tempdir.cleanup()

    @unittest.skipUnless(os.path.exists('/bin/yum'), 'No yum binary present on system')
    @unittest.skipIf('.' not in distro.version(), 'Distro version does not have a minor number')
    def test__install__default(self):
        installer = YumInstaller(dry_run=True)
        result = installer.install_dependencies()
        self.assertIn('python36', result)
